"""Import ntriples into KGTK format.
"""
from argparse import ArgumentParser, Namespace
import attr
import json
from pathlib import Path
import re
import sys
import typing

from kgtk.kgtkformat import KgtkFormat
from kgtk.io.kgtkreader import KgtkReader, KgtkReaderMode, KgtkReaderOptions
from kgtk.value.kgtkvalue import KgtkValue, KgtkValueFields
from kgtk.value.kgtkvalueoptions import KgtkValueOptions, DEFAULT_KGTK_VALUE_OPTIONS

@attr.s(slots=True, frozen=False)
class GroupedReader:
    reader: KgtkReader = attr.ib(validator=attr.validators.instance_of(KgtkReader))

    current_row: typing.Optional[typing.List[str]] = attr.ib(default=None)
    current_qnode: typing.Optional[str] = attr.ib(default=None)

    def fetch(self, qnode: str)->typing.List[typing.List[str]]:
        result: typing.List[typing.List[str]] = [ ]
        while True:
            if self.current_row is None:
                try:
                    self.current_row = self.reader.nextrow()
                except StopIteration:
                    return result

                if self.reader.node1_column_idx < 0:
                    raise ValueError("GroupedReader: no node1 index.")

                current_qnode: str = self.current_row[self.reader.node1_column_idx]
                if "-" in current_qnode:
                    current_qnode, _ = current_qnode.split("-", 1)
                self.current_qnode = current_qnode

            if self.current_qnode is None:
                raise ValueError("GroupedReader: self.current_qunode is unexpectedly missing.")

            if self.current_qnode < qnode:
                self.current_row = None
                continue

            if self.current_qnode > qnode:
                return result

            result.append(self.current_row)
            self.current_row = None
                
@attr.s(slots=True, frozen=False)
class ExportWikidata(KgtkFormat):
    # TODO: write a validator:
    node_file_path: Path = attr.ib(validator=attr.validators.instance_of(Path))
    edge_file_path: Path = attr.ib(validator=attr.validators.instance_of(Path))
    qualifier_file_path: Path = attr.ib(validator=attr.validators.instance_of(Path))

    output_file_path: Path = attr.ib(validator=attr.validators.instance_of(Path))

    reader_options: KgtkReaderOptions = attr.ib(validator=attr.validators.instance_of(KgtkReaderOptions))
    value_options: KgtkValueOptions = attr.ib(validator=attr.validators.instance_of(KgtkValueOptions), default=DEFAULT_KGTK_VALUE_OPTIONS)

    error_file: typing.TextIO = attr.ib(default=sys.stderr)
    verbose: bool = attr.ib(validator=attr.validators.instance_of(bool), default=False)
    very_verbose: bool = attr.ib(validator=attr.validators.instance_of(bool), default=False)

    node_alias_idx: int = attr.ib(default=-1)
    node_description_idx: int = attr.ib(default=-1)
    node_label_idx: int = attr.ib(default=-1)
    node_qnode_idx: int = attr.ib(default=-1)
    node_type_idx: int = attr.ib(default=-1)
    node_datatype_idx: int = attr.ib(default=-1)

    edge_id_idx: int = attr.ib(default=-1)
    edge_node1_idx: int = attr.ib(default=-1)
    edge_label_idx: int = attr.ib(default=-1)
    edge_node2_idx: int = attr.ib(default=-1)
    edge_rank_idx: int = attr.ib(default=-1)
    edge_wikidatatype_idx: int = attr.ib(default=-1)
    edge_claim_id_idx: int = attr.ib(default=-1)
    edge_val_type_idx: int = attr.ib(default=-1)
    edge_entity_type_idx: int = attr.ib(default=-1)
    edge_datahash_idx: int = attr.ib(default=-1)
    edge_precision_idx: int = attr.ib(default=-1)
    edge_calendar_idx: int = attr.ib(default=-1)

    qual_node1_idx: int = attr.ib(default=-1)
    qual_label_idx: int = attr.ib(default=-1)
    qual_node2_idx: int = attr.ib(default=-1)
    qual_wikidatatype_idx: int = attr.ib(default=-1)
    qual_val_type_idx: int = attr.ib(default=-1)
    qual_entity_type_idx: int = attr.ib(default=-1)
    qual_datahash_idx: int = attr.ib(default=-1)
    qual_precision_idx: int = attr.ib(default=-1)
    qual_calendar_idx: int = attr.ib(default=-1)

    def add_attr_to_map(self,
                        attr_map: typing.MutableMapping[str, typing.Mapping[str, str]],
                        attr: str,
                        who: str,
    ):
        kv: KgtkValue = KgtkValue(attr, options=self.value_options, parse_fields=False, error_file=self.error_file, verbose=self.verbose)
        if not kv.is_language_qualified_string(validate=True):
            raise ValueError("Invald attr %s for %s" % (attr, who))

        text: str
        language: str
        language_suffix: str
        text, language, language_suffix = KgtkFormat.destringify(kv.value)
        if len(language) == 0:
            raise ValueError("No attr language in %s for %s" % (attr, who))
        lang: str = language + language_suffix
        attr_map[lang] = {
                "language" : lang,
                "value": text
        }
            

    def add_attr(self,
                 result: typing.MutableMapping[str, typing.Any],
                 attr: str,
                 who: str,
    ):
        attr_map: typing.MutableMapping[str, typing.Mapping[str, str]]
        if who in result:
            attr_map = result[who]
        else:
            attr_map = dict()
            result[who] = attr_map

        self.add_attr_to_map(attr_map, attr, who)

    def build_attr_map(self,
                       result: typing.MutableMapping[str, typing.Any],
                       attr_list: str,
                       who: str,
    ):
        if len(attr_list) == 0:
            return
        
        attr_map: typing.MutableMapping[str, typing.Mapping[str, str]] = { }
        attr: str
        for attr in KgtkValue.split_list(attr_list):
            self.add_attr_to_map(attr_map, attr, who)

        if len(attr_map) > 0:
            result[who] = attr_map

    def add_qnode(self,
                 result: typing.MutableMapping[str, typing.Any],
                 qnode: str):
        result["id"] = qnode

    def add_type(self,
                 result: typing.MutableMapping[str, typing.Any],
                 node_type: str):
        result["type"] = node_type

    def add_datatype(self,
                 result: typing.MutableMapping[str, typing.Any],
                 node_type: str):
        if len(node_type) > 0:
            result["datatype"] = node_type

    def process_qnode_info(self,
                           qnode: str,
                           qnode_info: typing.List[str],
    )->typing.MutableMapping[str, typing.Any]:
        result: typing.MutableMapping[str, typing.Any] = dict()
        self.add_qnode(result, qnode)
        self.add_type(result, qnode_info[self.node_type_idx])
        self.add_datatype(result, qnode_info[self.node_datatype_idx])

        self.build_attr_map(result, qnode_info[self.node_label_idx], "labels")
        self.build_attr_map(result, qnode_info[self.node_description_idx], "descriptions")
        self.build_attr_map(result, qnode_info[self.node_alias_idx], "aliases")
        
        return result

    def build_qualifier_dict(self, qnode: str, qgr: GroupedReader)->typing.Mapping[str, typing.List[typing.List[str]]]:
        result: typing.MutableMapping[str, typing.List[typing.List[str]]] = dict()

        qualifiers: typing.List[typing.List[str]] = qgr.fetch(qnode)
        qualifier: typing.List[str]
        for qualifier in qualifiers:
            edge_id: str = qualifier[self.qual_node1_idx]
            if edge_id not in result:
                result[edge_id] = list()
            result[edge_id].append(qualifier)

        return result

    def add_sitelink(self,
                     result: typing.MutableMapping[str, typing.Any],
                     edge_id: str,
                     qualifier_rows: typing.List[typing.List[str]]):
        if "sitelinks" not in result:
            result["sitelinks"] = dict()
        sitelinks: typing.MutableMapping[str, typing.Mapping[str, typing.Union[str, typing.List[str]]]] = result["sitelinks"]

        site: str = ""
        title: str = ""
        badges: typing.List[str] = list()

        qualifier_row: typing.List[str]
        for qualifier_row in qualifier_rows:
            label: str = qualifier_row[self.qual_label_idx]

            if label == "site":
                site = qualifier_row[self.qual_node2_idx]

            elif label == "title":
                title = KgtkFormat.unstringify(qualifier_row[self.qual_node2_idx])

            elif label == "badge":
                badges.append(qualifier_row[self.qual_node2_idx])

        if len(site) == 0:
            # TODO: give a better error message.
            raise ValueError("Missing sitelink site for %s" % edge_id)

        if len(title) == 0:
            # TODO: give a better error message.
            raise ValueError("Missing sitelink title for %s" % edge_id)

        sitelinks[site] = {
            "site": site,
            "title": title,
            "badges": badges
        }
            

    def process_edge_datavalue(self,
                               value: str,
                               edge_row: typing.List[str],
                               datatype: str):
        datavalue: typing.MutableMapping[str, typing.Union[str, typing.Mapping[str, typing.Optional[typing.Union[str, int, float]]]]] = dict()
        datavalue["type"] = edge_row[self.edge_val_type_idx]

        valuemap: typing.MutableMapping[str, typing.Optional[typing.Union[str, int, float]]] = dict()
        datavalue["value"] = valuemap

        entity_type: str = edge_row[self.edge_entity_type_idx]
        if len(entity_type) > 0:
            valuemap["entity-type"] = entity_type
            valuemap["id"] = value

            # TODO: Is this the right thing to do?
            numeric_id: str = value[1:]
            if "-" in numeric_id:
                numeric_id = numeric_id[:numeric_id.index("-")]
            valuemap["numeric-id"] = int(numeric_id)
            return datavalue

        kv = KgtkValue(value, options=self.value_options, parse_fields=True, error_file=self.error_file, verbose=self.verbose)
        if not kv.validate():
            # raise ValueError("Invalid KGTK value '%s'" % value)
            print("Warning: Invalid KGTK value '%s'" % value, file=self.error_file, flush=True)
        if kv.fields is None:
            raise ValueError("KGTK value '%s' is missing fields." %value)

        if kv.is_number():
            if kv.fields.numberstr is None:
                raise ValueError("number is missing numberstr.")
                
            valuemap["amount"] = kv.fields.numberstr # TODO: add plus sign
            valuemap["unit"] = "1"
            return datavalue

        if kv.is_quantity():
            if kv.fields.numberstr is None:
                raise ValueError("quantity is missing numberstr.")
            valuemap["amount"] = kv.fields.numberstr # TODO: add plus sign

            if kv.fields.units_node is None:
                # TODO: research this further.
                #
                # raise ValueError("quantity is missing units_node for %s." % value)
                valuemap["init"] = "undefined"
            else:
                valuemap["unit"] = "http://www.wikidata.org/entity/" + kv.fields.units_node

            if kv.fields.low_tolerancestr is not None and len(kv.fields.low_tolerancestr) > 0:
                valuemap["lowerBound"] = kv.fields.low_tolerancestr # TODO: add plus sign

            if kv.fields.high_tolerancestr is not None and len(kv.fields.high_tolerancestr) > 0:
                valuemap["higherBound"] = kv.fields.high_tolerancestr # TODO: add plus sign
            return datavalue

        if kv.is_language_qualified_string():
            text: str
            language: str
            language_suffix: str
            text, language, language_suffix = KgtkFormat.destringify(value) # TODO: KgtkValue should do this to text
            language += language_suffix
            valuemap["text"] = text
            valuemap["language"] = language
            return datavalue
        
        if kv.is_string():
            valuemap["type"] = "string"
            valuemap["value"] = KgtkFormat.unstringify(value) # TODO: KgtkValue should do this to text
            return datavalue

        if kv.is_date_and_times():
            if kv.fields.zonestr is None:
                raise ValueError("timezone is missing.")
            if kv.fields.zonestr != "Z":
                raise ValueError("Only Z-time is supported.")

            if kv.fields.date_and_time is None:
                raise ValueError("date_and_time is missing.")
            valuemap["time"] = kv.fields.date_and_time
            valuemap["timezone"] = 0
            valuemap["before"] = 0
            valuemap["after"] = 0
        
            if kv.fields.precision is None:
                raise ValueError("date_and_time precision is missing.")
            valuemap["precision"] = kv.fields.precision

            valuemap["calendarmodel"] = "http://www.wikidata.org/entity/" + edge_row[self.edge_calendar_idx]
            return datavalue

        if kv.is_location_coordinates:
            if kv.fields.latitude is None:
                raise ValueError("latitude is missing")
            valuemap["latitude"] = kv.fields.latitude

            if kv.fields.longitude is None:
                raise ValueError("longitude is missing")
            valuemap["longitude"] = kv.fields.longitude

            valuemap["altitide"] = None # deprecated

            # TODO: Validate that it's OK to have location coordinates without precision.
            precision: str = edge_row[self.edge_precision_idx]
            if len(precision) > 0:
                try:
                    valuemap["precision"] = float(edge_row[self.edge_precision_idx])
                except ValueError:
                    print("Invalid precision '%s'" % precision, file=self.error_file, flush=True)

            valuemap["globe"] = "http://www.wikidata.org/entity/Q2"
            return datavalue


        # Default: treat as string.
        valuemap["type"] = "string"
        valuemap["value"] = KgtkFormat.unstringify(value) # TODO: KgtkValue should do this to text
        return datavalue

    def process_qual_datavalue(self,
                               value: str,
                               qual_row: typing.List[str],
                               datatype: str):
        datavalue: typing.MutableMapping[str, typing.Union[str, typing.Mapping[str, typing.Optional[typing.Union[str, int, float]]]]] = dict()
        datavalue["type"] = qual_row[self.qual_val_type_idx]

        valuemap: typing.MutableMapping[str, typing.Optional[typing.Union[str, int, float]]] = dict()
        datavalue["value"] = valuemap

        entity_type: str = qual_row[self.qual_entity_type_idx]
        if len(entity_type) > 0:
            valuemap["entity-type"] = entity_type
            valuemap["id"] = value

            # TODO: Is this the right thing to do for Q16097-F1?
            numeric_id: str = value[1:]
            if "-" in numeric_id:
                numeric_id = numeric_id[:numeric_id.index("-")]
            valuemap["numeric-id"] = int(numeric_id)
            return datavalue

        kv = KgtkValue(value, options=self.value_options, parse_fields=True, error_file=self.error_file, verbose=self.verbose)
        if not kv.validate():
            # raise ValueError("Invalid KGTK value '%s'" % value)
            print("Warning: Invalid KGTK value '%s'" % value, file=self.error_file, flush=True)
        if kv.fields is None:
            raise ValueError("KGTK value %s is missing fields." % value)

        if kv.is_number():
            if kv.fields.numberstr is None:
                raise ValueError("number is missing numberstr for %s." % value)
                
            valuemap["amount"] = kv.fields.numberstr # TODO: add plus sign
            valuemap["unit"] = "1"
            return datavalue

        if kv.is_quantity():
            if kv.fields.numberstr is None:
                raise ValueError("quantity is missing numberstr for %s." % value)
            valuemap["amount"] = kv.fields.numberstr # TODO: add plus sign

            if kv.fields.units_node is None:
                # TODO: Research this further.  Why did we get here?  Is it because import_wikidata
                # dropped the units?
                #
                # raise ValueError("quantity is missing units_node for %s in: %s" % (value, " ".join(qual_row)))
                valuemap["unit"] = "undefined"
            else:
                valuemap["unit"] = "http://www.wikidata.org/entity/" + kv.fields.units_node

            if kv.fields.low_tolerancestr is not None and len(kv.fields.low_tolerancestr) > 0:
                valuemap["lowerBound"] = kv.fields.low_tolerancestr # TODO: add plus sign

            if kv.fields.high_tolerancestr is not None and len(kv.fields.high_tolerancestr) > 0:
                valuemap["higherBound"] = kv.fields.high_tolerancestr # TODO: add plus sign
            return datavalue

        if kv.is_language_qualified_string():
            text: str
            language: str
            language_suffix: str
            text, language, language_suffix = KgtkFormat.destringify(value) # TODO: KgtkValue should do this to text
            language += language_suffix
            valuemap["text"] = text
            valuemap["language"] = language
            return datavalue
        
        if kv.is_string():
            valuemap["type"] = "string"
            valuemap["value"] = KgtkFormat.unstringify(value) # TODO: KgtkValue should do this to text
            return datavalue

        if kv.is_date_and_times():
            if kv.fields.zonestr is None:
                raise ValueError("timezone is missing from %s." % value)
            if kv.fields.zonestr != "Z":
                raise ValueError("Only Z-time is supported, error in %s." % value)

            if kv.fields.date_and_time is None:
                raise ValueError("date_and_time is missing from %s." % value)
            valuemap["time"] = kv.fields.date_and_time
            valuemap["timezone"] = 0
            valuemap["before"] = 0
            valuemap["after"] = 0
        
            if kv.fields.precision is None:
                raise ValueError("date_and_time precision is missing from %s." % value)
            valuemap["precision"] = kv.fields.precision

            valuemap["calendarmodel"] = "http://www.wikidata.org/entity/" + qual_row[self.qual_calendar_idx]
            return datavalue

        if kv.is_location_coordinates():
            if kv.fields.latitude is None:
                raise ValueError("latitude is missing from %s" % value)
            valuemap["latitude"] = kv.fields.latitude

            if kv.fields.longitude is None:
                raise ValueError("longitude is missing from %s" % value)
            valuemap["longitude"] = kv.fields.longitude

            valuemap["altitide"] = None # deprecated

            valuemap["precision"] = float(qual_row[self.qual_precision_idx])

            valuemap["globe"] = "http://www.wikidata.org/entity/Q2"
            return datavalue


        # Default: convert the symbol to a string.
        valuemap["type"] = "string"
        valuemap["value"] = KgtkFormat.unstringify('"' + value + '"') # TODO: KgtkValue should do this to text
        return datavalue

    def process_qnode_edge_qualifier(self,
                           statement: typing.MutableMapping[str, typing.Any],
                           edge_id: str,
                           qualifier_row: typing.List[str]):
        if "qualifiers" not in statement:
            statement["qualifiers"] = dict()
        qualifiers = statement["qualifiers"]

        prop: str = qualifier_row[self.qual_label_idx]
        if prop not in qualifiers:
            qualifiers[prop] = list()
        proplist: typing.List[typing.Mapping[str, typing.Any]] = qualifiers[prop]

        qualifier: typing.MutableMapping[str, typing.Any] = dict()
        proplist.append(qualifier)

        qualifier["property"] = prop

        datatype: str = qualifier_row[self.qual_wikidatatype_idx]
        qualifier["datatype"] = datatype

        datahash: str = qualifier_row[self.qual_datahash_idx]
        if len(datahash) > 0:
            qualifier["hash"] = KgtkFormat.unstringify(datahash)

        value: str = qualifier_row[self.qual_node2_idx]
        if value == "somevalue":
            qualifier["snaktype"] = "somevalue"
        elif value == "novalue":
            qualifier["snaktype"] = "novalue"
        else:
            qualifier["datavalue"] = self.process_qual_datavalue(value, qualifier_row, datatype)

    def process_qnode_edge_qualifiers(self,
                                      statement: typing.MutableMapping[str, typing.Any],
                                      edge_id: str,
                                      qualifier_rows: typing.List[typing.List[str]]):
        qualifier_row: typing.List[str]
        for qualifier_row in qualifier_rows:
            self.process_qnode_edge_qualifier(statement, edge_id, qualifier_row)

    def process_qnode_edge(self,
                           result: typing.MutableMapping[str, typing.Any],
                           qnode: str,
                           prop: str,
                           edge_id: str,
                           edge_row: typing.List[str],
                           qualifier_rows: typing.Optional[typing.List[typing.List[str]]]):
        if "claims" not in result:
            result["claims"] = dict()
        claims = result["claims"]

        if prop not in claims:
            claims[prop] = list()
        proplist: typing.List[typing.Mapping[str, typing.Any]] = claims[prop]

        try:
            statement: typing.MutableMapping[str, typing.Any] = dict()
            
            # Fill in the top-level attributes:
            statement["type"] = "statement" # We hope this is correct.
            statement["id"] = edge_row[self.edge_claim_id_idx]
            statement["rank"] = edge_row[self.edge_rank_idx]

            mainsnak: typing.MutableMapping[str, typing.Any] = dict()
            statement["mainsnak"] = mainsnak

            mainsnak["property"] = prop

            datatype: str = edge_row[self.edge_wikidatatype_idx]
            mainsnak["datatype"] = datatype

            value: str = edge_row[self.edge_node2_idx]
            if value == "somevalue":
                mainsnak["snaktype"] = "somevalue"
            elif value == "novalue":
                mainsnak["snaktype"] = "novalue"
            else:
                mainsnak["snaktype"] = "value"
                mainsnak["datavalue"] = self.process_edge_datavalue(value, edge_row, datatype)

            if qualifier_rows is not None and len(qualifier_rows) > 0:
                self.process_qnode_edge_qualifiers(result, edge_id, qualifier_rows)

            proplist.append(statement)
        
        except ValueError as e:
            print("Error, skipping statement: %s" % str(e), file=self.error_file, flush=True)

    def process_qnode_edges(self,
                            result: typing.MutableMapping[str, typing.Any],
                            qnode: str,
                            edges: typing.List[typing.List[str]],
                            qualifier_dict: typing.Mapping[str, typing.List[typing.List[str]]]):

        # In case we're not using a node file.
        self.add_qnode(result, qnode)

        edge_id: str
        qualifier_rows: typing.Optional[typing.List[typing.List[str]]]

        edge_row: typing.List[str]
        for edge_row in edges:
            edge_label: str = edge_row[self.edge_label_idx]

            if edge_label == "alias":
                self.add_attr(result, edge_row[self.edge_node2_idx], "alias")

            elif edge_label == "description":
                self.add_attr(result, edge_row[self.edge_node2_idx], "description")
            
            elif edge_label == "label":
                self.add_attr(result, edge_row[self.edge_node2_idx], "label")
            
            elif edge_label == "type":
                self.add_type(result, edge_row[self.edge_node2_idx])

            elif edge_label == "datatype":
                self.add_datatype(result, edge_row[self.edge_node2_idx])

            elif edge_label in ("wikipedia_sitelink", "addl_wikipedia_sitelink"):
                edge_id = edge_row[self.edge_id_idx]
                qualifier_rows = qualifier_dict.get(edge_id)
                if qualifier_rows is not None and len(qualifier_rows) > 0:
                    self.add_sitelink(result, edge_id, qualifier_rows)

            else:
                edge_id = edge_row[self.edge_id_idx]
                qualifier_rows = qualifier_dict.get(edge_id)
                self.process_qnode_edge(result, qnode, edge_label, edge_id, edge_row, qualifier_rows)
                            

    def process_qnode(self,
                      qnode_info: typing.List[str],
                      egr: GroupedReader,
                      qgr: GroupedReader,
    )->typing.Mapping[str, typing.Any]:
        qnode: str = qnode_info[self.node_qnode_idx]
        if self.very_verbose:
            print("Processing qnode %s" % qnode)

        result: typing.MutableMapping[str, typing.Any] =  self.process_qnode_info(qnode, qnode_info)

        edges: typing.List[typing.List[str]] = egr.fetch(qnode)
        qualifier_dict: typing.Mapping[str, typing.List[typing.List[str]]] = self.build_qualifier_dict(qnode, qgr)

        self.process_qnode_edges(result, qnode, edges, qualifier_dict)
        
        return result
            
    def get_required_columns(self, nr: KgtkReader, er: KgtkReader, qr: KgtkReader):
        self.node_qnode_idx = nr.id_column_idx
        if self.node_qnode_idx < 0:
            raise ValueError("The node file is missing an ID column.")

        self.node_label_idx = nr.column_name_map.get("label", -1)
        if self.node_label_idx < 0:
            raise ValueError("The node file is missing a label column.")

        self.node_type_idx = nr.column_name_map.get("type", -1)
        if self.node_type_idx < 0:
            raise ValueError("The node file is missing a type column.")

        self.node_datatype_idx = nr.column_name_map.get("datatype", -1)
        if self.node_datatype_idx < 0:
            raise ValueError("The node file is missing a datatype column.")

        self.node_description_idx = nr.column_name_map.get("description", -1)
        if self.node_description_idx < 0:
            raise ValueError("The node file is missing a description column.")

        self.node_alias_idx = nr.column_name_map.get("alias", -1)
        if self.node_alias_idx < 0:
            raise ValueError("The node file is missing a alias column.")


        self.edge_id_idx = er.id_column_idx
        if self.edge_id_idx < 0:
            raise ValueError("The edge file does not have a id column.")

        self.edge_node1_idx = er.node1_column_idx
        if self.edge_node1_idx < 0:
            raise ValueError("The edge file does not have a node1 column.")

        self.edge_label_idx = er.label_column_idx
        if self.edge_label_idx < 0:
            raise ValueError("The edge file does not have a label column.")

        self.edge_node2_idx = er.node2_column_idx
        if self.edge_node2_idx < 0:
            raise ValueError("The edge file does not have a node2 column.")

        self.edge_rank_idx = er.column_name_map.get("rank", -1)
        if self.edge_rank_idx < 0:
            raise ValueError("The edge file does not have a rank column.")

        self.edge_wikidatatype_idx = er.column_name_map.get("node2;wikidatatype", -1)
        if self.edge_wikidatatype_idx < 0:
            raise ValueError("The edge file does not have a node2:wikidatatype column.")

        self.edge_claim_id_idx = er.column_name_map.get("claim_id", -1)
        if self.edge_claim_id_idx < 0:
            raise ValueError("The edge file does not have a claim_id column.")

        self.edge_val_type_idx = er.column_name_map.get("val_type", -1)
        if self.edge_val_type_idx < 0:
            raise ValueError("The edge file does not have a val_type column.")

        self.edge_entity_type_idx = er.column_name_map.get("entity_type", -1)
        if self.edge_entity_type_idx < 0:
            raise ValueError("The edge file does not have an entity_type column.")

        self.edge_datahash_idx = er.column_name_map.get("datahash", -1)
        if self.edge_datahash_idx < 0:
            raise ValueError("The edge file does not have a datahash column.")

        self.edge_precision_idx = er.column_name_map.get("precision", -1)
        if self.edge_precision_idx < 0:
            raise ValueError("The edge file does not have a precision column.")

        self.edge_calendar_idx = er.column_name_map.get("calendar", -1)
        if self.edge_calendar_idx < 0:
            raise ValueError("The edge file does not have a calendar column.")


        self.qual_node1_idx = qr.node1_column_idx
        if self.qual_node1_idx < 0:
            raise ValueError("The qualifier file does not have a node1 column.")

        self.qual_label_idx = qr.label_column_idx
        if self.qual_label_idx < 0:
            raise ValueError("The qualifier file does not have a label column.")

        self.qual_node2_idx = qr.node2_column_idx
        if self.qual_node2_idx < 0:
            raise ValueError("The qualifier file does not have a node2 column.")

        self.qual_wikidatatype_idx = qr.column_name_map.get("node2;wikidatatype", -1)
        if self.qual_wikidatatype_idx < 0:
            raise ValueError("The qualifier file does not have a node2:wikidatatype column.")

        self.qual_val_type_idx = qr.column_name_map.get("val_type", -1)
        if self.qual_val_type_idx < 0:
            raise ValueError("The qual file does not have a val_type column.")

        self.qual_entity_type_idx = qr.column_name_map.get("entity_type", -1)
        if self.qual_entity_type_idx < 0:
            raise ValueError("The qual file does not have an entity_type column.")

        self.qual_datahash_idx = qr.column_name_map.get("datahash", -1)
        if self.qual_datahash_idx < 0:
            raise ValueError("The qual file does not have a datahash column.")

        self.qual_precision_idx = qr.column_name_map.get("precision", -1)
        if self.qual_precision_idx < 0:
            raise ValueError("The qual file does not have a precision column.")

        self.qual_calendar_idx = qr.column_name_map.get("calendar", -1)
        if self.qual_calendar_idx < 0:
            raise ValueError("The qual file does not have a calendar column.")

    def process(self):

        if self.verbose:
            print("Opening output file %s" % str(self.output_file_path), file=self.error_file, flush=True)
        outfile: typing.TextIO = open(self.output_file_path, "wt")
        outfile.write("[")

        if self.verbose:
            print("Opening the node file: %s" % str(self.node_file_path), file=self.error_file, flush=True)
        nr: KgtkReader = KgtkReader.open(self.node_file_path,
                                         error_file=self.error_file,
                                         options=self.reader_options,
                                         value_options = self.value_options,
                                         verbose=self.verbose,
                                         very_verbose=self.very_verbose,
        )

        if self.verbose:
            print("Opening the edge file: %s" % str(self.edge_file_path), file=self.error_file, flush=True)
        er: KgtkReader = KgtkReader.open(self.edge_file_path,
                                         error_file=self.error_file,
                                         options=self.reader_options,
                                         value_options = self.value_options,
                                         verbose=self.verbose,
                                         very_verbose=self.very_verbose,
        )
        egr: GroupedReader = GroupedReader(reader=er)

        if self.verbose:
            print("Opening the qualifier file: %s" % str(self.qualifier_file_path), file=self.error_file, flush=True)
        qr: KgtkReader = KgtkReader.open(self.qualifier_file_path,
                                         error_file=self.error_file,
                                         options=self.reader_options,
                                         value_options = self.value_options,
                                         verbose=self.verbose,
                                         very_verbose=self.very_verbose,
        )
        qgr: GroupedReader = GroupedReader(reader=qr)

        self.get_required_columns(nr, er, qr)

        qnode_count: int = 0
        first: bool = True
        qnode_info: typing.List[str]
        for qnode_info in nr:
            result: typing.Mapping[str, typing.Any] = self.process_qnode(qnode_info,
                                                                         egr=egr,
                                                                         qgr=qgr)
            if first:
                first = False
                outfile.write("\n")
            else:
                outfile.write(",\n")
            outfile.write(json.dumps(result, indent=None, separators=(',', ':'), sort_keys=True))

        outfile.write("\n]\n")
        outfile.close()

def main():
    """
    Test the Wikidata Exporter.
    """
    pass
    
if __name__ == "__main__":
    main()



"""
'VectorFunction's to support Kypher queries and computations over vectors.
"""

import numpy as np

from   kgtk.exceptions import KGTKException
from   kgtk.kypher.utils import *
import kgtk.kypher.indexspec as ispec
from   kgtk.kypher.functions import SqlFunction, VirtualGraphFunction
from   kgtk.kypher.vecstore import MasterVectorStore, InlineVectorStore
import kgtk.kypher.parser as parser


### NOTES

# To properly access and efficiently compute with indexed vector data we have to support
# two basic operations:
#
# (1) graph variable access: when a graph variable corresponding to a vector column is
# referenced in a Kypher query (e.g., in a 'where' or 'return'-clause), it has to
# generate an appropriate vector object.  For columns indexed via an 'inline'-store this
# is automatic, the value will simply be the byte string representing the BLOB value of
# the encoded vector column.  For external array-based stores ('numpy' and 'hd5') we have
# to substitute the variable reference with a function call that brings in the vector
# object from the secondary store.  This is done by creating calls to '_kvec_get_vector'
# which takes the 'rowid' of the graph column as its argument and retrieves the bytes
# from the corresponding store based on that index (during import each vector at 'rowid=N'
# will be stored at position 'store[N-1]' in the external store, since row IDs are 1-based).
# The Kypher query translator calls 'SqlStore.vector_column_to_sql' to generate these
# accesses, which in turn will call out to the appropriate graph table.column vector store.
# The 'KvecGetVector' function below will create a call-specific function object that
# has an efficient cache for the required vector store handle built-in.
#
# (2) function calls processing or returning vectors: functions called by the SQLite query
# engine can only be passed and return literal data types and binary BLOB values (represented
# as byte strings in the Python API).  This means we cannot pass Python objects such as NumPy
# vectors from one function to another (e.g., the result of a '_kvec_get_vector' call).
# This means we have to either convert a NumPy vector to its underlying bytes before passing
# or returning, or we create special-purpose optimized translations for certain situations
# (in the current implementation we use both schemes).  For example, a call of a vector
# function in a Kypher query such as, for example,
#    kvec_dot_product(v1, v2)
# can be translated into this if both variables correspond to (compatible) inline stores:
#    kvec_dot_product_inline(graph_1_c1.node2, graph_1_c2.node2)
# Alternatively, if the variables correspond to a NumPy store, this translation can be used:
#    kvec_dot_product_array(graph_1_c1.rowid, graph_1_c2.rowid)
# A generic call can first create bytes and then pass them to the underlying function like so:
#    kvec_dot_product_generic(_kvec_get_vector(graph_1_c1.rowid), _kvec_get_vector(graph_1_c2.rowid))
# The function object 'DotProduct' handles the context-specific translations and will try to
# use optimized calls where possible or a generic call if optimization can't be done.  It will
# also create call-specific custom function pointers (closures) that encapsulate all the
# necessary state and cache information (e.g., store handles, vector dtypes, etc.) so that
# they do not need to be passed in or determined at run time for optimal efficiency.
#
# The function objects below can do further query-translation-time optimizations.  For example,
# 'CosineSimilarity' can check whether the corresponding vector columns were appropriately
# normalized, and, if so, can subsitute a cheaper dot-product computation.


class VectorFunction(SqlFunction):
    """General function class supporting vector computations.
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.uniquify = True
        self.xstore = None     # used to hold the vector store corresponding to the first x argument
        self.ystore = None     # used to hold the vector store corresponding to the second y argument
        self.optimize = False  # internal switch to flag whether to produce optimized or generic code

    def get_argument_vector_store(self, query, arg, state):
        """If 'arg' is a column variable that can be linked to a vector store,
        return the associated vector store object, otherwise return None.
        As the second return value, return the optimized reference that can
        be used for the vector for this type of vector store.
        """
        # TO DO: figure out whether we need to worry about join-variables here
        # that can't be linked unambiguously to a table, but then joins on 
        # vector columns don't really make sense, do they?
        if isinstance(arg, parser.Variable):
            graph, column, sql = query.variable_to_sql(arg, state)
            table = query.graph_alias_to_graph(graph)
            master = self.store.get_vector_store()
            if master.has_vector_store(table, column):
                vstore = master.get_vector_store(table, column)
                vref = vstore.vector_column_to_reference_sql(table_alias=graph)
                return vstore, vref
        return None, None

    def get_call_context(self):
        """Define a set of closure variables needed to define optimized vector operation code.
        If uniquify=True (the default), these will be specific to a single call location in a 
        query and not be shared across calls or queries.  NOTE: the default dtypes here are
        intended for the generic call case where we might not know, for example, in case an
        argument is provided through another function call.
        """
        sql_store, xstore, ystore = self.store, self.xstore, self.ystore
        xtable, xcolumn, xdtype = None, None, np.float32
        xcache = [None, None]
        if xstore is not None:
            xtable = xstore.table
            xcolumn = xstore.column
            xdtype = xstore.get_vector_dtype()
        ytable, ycolumn, ydtype = None, None, np.float32
        ycache = [None, None]
        if ystore is not None:
            ytable = ystore.table
            ycolumn = ystore.column
            ydtype = ystore.get_vector_dtype()
        return sql_store, xtable, xcolumn, xdtype, xcache, ytable, ycolumn, ydtype, ycache

    def translate_two_arg_call_to_optimized_sql(self, query, expr, state):
        args = expr.args
        if len(args) != 2:
            raise Exception("Illegal number of arguments to '{self.get_name()}'")
        self.xstore, xref = self.get_argument_vector_store(query, args[0], state)
        self.ystore, yref = self.get_argument_vector_store(query, args[1], state)
        self.optimize = self.xstore and self.ystore and type(self.xstore) == type(self.ystore)
        if self.optimize:
            # we produce an optimized call operating on vector references:
            args = [xref, yref]
        else:
            # otherwise, each arg is assumed to yield a byte-string encoding a vector:
            args = [query.expression_to_sql(arg, state) for arg in args]
        self.load()
        return f'{self.get_name()}({", ".join(args)})'

    def translate_two_arg_call_to_generic_sql(self, query, expr, state):
        args = expr.args
        if len(args) != 2:
            raise Exception("Illegal number of arguments to '{self.get_name()}'")
        self.xstore, xref = self.get_argument_vector_store(query, args[0], state)
        self.ystore, yref = self.get_argument_vector_store(query, args[1], state)
        self.optimize = False
        # each arg is assumed to yield a byte-string encoding a vector:
        args = [query.expression_to_sql(arg, state) for arg in args]
        self.load()
        return f'{self.get_name()}({", ".join(args)})'


class KvecGetVector(VectorFunction):
    """Access the vector corresponding to a vector-indexed column variable as a string of bytes.
    This is an internal accessor needed to retrieve vectors from Numpy and HD5 vector stores.
    """

    def get_code(self):
        if self.code is None:
            if self.xstore is None:
                raise KGTKException(f"missing vector store for call to '{self.get_name()}'")
            elif isinstance(self.xstore, InlineVectorStore):
                def _kvec_get_vector(vector):
                    return vector
            else:
                sstore, xtab, xcol, xdtype, xcache, _, _, _, _ = self.get_call_context()
                def _kvec_get_vector(rowid):
                    vs = xcache[0]
                    if vs is None:
                        # this will initialize the cache to the proper store handle upon the first call:
                        vs = MasterVectorStore.get_vector_store_as_array(sstore, xtab, xcol, cache=xcache)
                    return vs[rowid - 1].tobytes()
            self.code = _kvec_get_vector
        return self.code

    def load(self):
        if isinstance(self.xstore, InlineVectorStore):
            # not needed for inline stores:
            return
        super().load()

KvecGetVector('_kvec_get_vector', num_params=1, deterministic=True).define()


class DotProduct(VectorFunction):
    """Compute the dot product between two vectors x and y.
    """

    # The implementation below is somewhat complex and shows the different kinds of optimizations
    # that are possible.  However, the generic call scheme is identical with the optimized scheme
    # for inline-store vectors, and will be just slightly slower for array-store vectors compared
    # to their optimized versions.  So a streamlined version could eliminate all optimizations and
    # simply implement the generic call scheme only, with just a very minor performance penalty.
    # However, we keep this for now as documentation for what can be done.
    
    def translate_call_to_sql(self, query, expr, state):
        return self.translate_two_arg_call_to_optimized_sql(query, expr, state)
        # for measuring improvements coming from optimized translations:
        #return self.translate_two_arg_call_to_generic_sql(query, expr, state)

    def get_optimized_code_inline(self):
        """Return optimized code operating on two arguments corresponding to inline vector stores.
        Since those will be byte strings coming directly from the corresponding graph table columns,
        no specific vector access is required and both 'x' and 'y' will be bound to bytes.
        """
        sstore, xtab, xcol, xdtype, xcache, ytab, ycol, ydtype, ycache = self.get_call_context()
        def _dot_product(x, y):
            # 'frombuffer' calls are very fast, O(1):
            vx = np.frombuffer(x, dtype=xdtype)
            vy = np.frombuffer(y, dtype=ydtype)
            # make sure we don't return numpy floats:
            return float(np.dot(vx, vy))
        return _dot_product

    def get_optimized_code_array_v1(self):
        """Return optimized code operating on two arguments corresponding to vector stores that
        can be accessed via array handles (numpy and hd5).  Arguments will be rowid's which are
        used to access the corresponding vectors from the array handles.
        """
        sstore, xtab, xcol, xdtype, xcache, ytab, ycol, ydtype, ycache = self.get_call_context()
        def _dot_product(xrowid, yrowid):
            # this is about 2x slower for the numpy store compared to inline, hd5 is much much slower than
            # that due to the individual array accesses causing significant per-call overhead; we tried
            # various tweaks on this for numpy, but this is about as fast as we can make it for now:
            xvs = xcache[0]
            if xvs is None:
                # this will initialize caches to the proper store handles upon the first call:
                xvs = MasterVectorStore.get_vector_store_as_array(sstore, xtab, xcol, cache=xcache)
                yvs = MasterVectorStore.get_vector_store_as_array(sstore, ytab, ycol, cache=ycache)
            # make sure we don't return numpy floats:
            return float(np.dot(xvs[xrowid - 1], ycache[0][yrowid - 1]))
        return _dot_product

    def get_optimized_code_array_v2(self):
        """Variant of 'get_optimized_code_array_v1' that copies vectors into local cache vector buffers
        to avoid any np array allocations.  This is very slightly faster than v1 but probably doesn't
        warrant the extra complexity.
        """
        sstore, xtab, xcol, xdtype, xcache, ytab, ycol, ydtype, ycache = self.get_call_context()
        def _dot_product(xrowid, yrowid):
            xv = xcache[1]
            yv = ycache[1]
            if xv is None:
                # this will initialize caches to the proper store handles upon the first call:
                MasterVectorStore.get_vector_store_as_array(sstore, xtab, xcol, cache=xcache)
                MasterVectorStore.get_vector_store_as_array(sstore, ytab, ycol, cache=ycache)
                xv = np.zeros(len(xcache[0][xrowid - 1]), dtype=xdtype)
                yv = np.zeros(len(ycache[0][yrowid - 1]), dtype=ydtype)
                xcache[1], ycache[1] = xv, yv
            xv[:] = xcache[0][xrowid - 1]
            yv[:] = ycache[0][yrowid - 1]
            # make sure we don't return numpy floats:
            return float(np.dot(xv, yv))
        return _dot_product

    def get_generic_code(self):
        """Generic calls assume byte strings as inputs which either come directly from the DB
        or via '_kvec_get_vector' calls.  For now this is the same as for inline stores, but
        we might change that down the road to deal with the potentially lossy dtype inference,
        in case a vector is passed through an intermediary function call.
        """
        return self.get_optimized_code_inline()
        
    def get_code(self):
        if self.code is None:
            if self.optimize:
                # we can and should produce an optimized call:
                if isinstance(self.xstore, InlineVectorStore):
                    self.code = self.get_optimized_code_inline()
                    # embellish the name for the benefit of --debug:
                    self.name += '_inline'
                # this case is currently only used if the two array stores are of the same type,
                # even though they would also work for a mix of numpy and hd5 stores:
                else:
                    self.code = self.get_optimized_code_array_v2()
                    self.name += '_array_v2'
            else:
                self.code = self.get_generic_code()
                self.name += '_generic'
        return self.code

DotProduct('kvec_dot', num_params=2, deterministic=True).define()
DotProduct('kvec_dot_product', num_params=2, deterministic=True).define()


class CosineSimilarity(DotProduct):
    """Compute cosine similarity between two vectors x and y.
    Use faster dot-product if both vectors are appropriately normalized.
    """

    def have_l2_normalized_vectors(self):
        """Return True if both vectors are L2-normalized.
        """
        xnorm = self.xstore.get_vector_norm() if self.xstore is not None else False
        ynorm = self.ystore.get_vector_norm() if self.ystore is not None else False
        return xnorm == ispec.VectorIndex.NORM_L2 and ynorm == ispec.VectorIndex.NORM_L2

    def have_unnormalized_vectors(self):
        """Return True if neither vector is (known to be) normalized.
        """
        xnorm = self.xstore.get_vector_norm() if self.xstore is not None else False
        ynorm = self.ystore.get_vector_norm() if self.ystore is not None else False
        return not xnorm and not ynorm

    def translate_call_to_sql(self, query, expr, state):
        if len(expr.args) != 2:
            raise Exception("Illegal number of arguments to '{self.get_name()}'")
        self.xstore, _ = self.get_argument_vector_store(query, expr.args[0], state)
        self.ystore, _ = self.get_argument_vector_store(query, expr.args[1], state)
        if self.have_l2_normalized_vectors():
            # both vectors are normalized with L2, which means we can simply run
            # the dot product and we should produce an optimized translation:
            return self.translate_two_arg_call_to_optimized_sql(query, expr, state)
        else:
            return self.translate_two_arg_call_to_generic_sql(query, expr, state)

    def get_generic_unnormalized_vector_code(self):
        sstore, xtab, xcol, xdtype, xcache, ytab, ycol, ydtype, ycache = self.get_call_context()
        def _cosine_similarity(x, y):
            # 'frombuffer' calls are very fast, O(1):
            vx = np.frombuffer(x, dtype=xdtype)
            vy = np.frombuffer(y, dtype=ydtype)
            norm = np.linalg.norm(vx) * np.linalg.norm(vy)
            # make sure we don't return numpy floats:
            return float(np.dot(vx, vy) / norm)
        return _cosine_similarity
    
    def get_code(self):
        if self.code is None:
            if self.have_l2_normalized_vectors():
                # both vectors are normalized with L2, simply run dot product,
                # we adapt the name for the benefit of --debug:
                self.name = 'kvec_dot_product'
                self.code = super().get_code()
            elif self.have_unnormalized_vectors():
                # neither vector is (known to be) normalized:
                self.code = self.get_generic_unnormalized_vector_code()
            else:
                raise KGTKException(f"cannot yet handle non-L2 or mix of norms in call to '{self.get_name()}'")
        return self.code

CosineSimilarity('kvec_cos_sim', num_params=2, deterministic=True).define()
CosineSimilarity('kvec_cosine_similarity', num_params=2, deterministic=True).define()


class TopKCosineSimilarity(VirtualGraphFunction, VectorFunction):
    """Find the top-k vectors with highest cosine similarity to a source vector.
    Requires the availability of an appropriate nearest-neighbor index on the
    associated vector store.
    """

    # parameters needed by the 'VirtualTableFunction' API:
    params = ['node1', 'k', 'nprobe']
    columns = ['label', 'node2', 'vid', 'vrowid', 'vector', 'similarity']
    name = 'kvec_topk_cosine_similarity'

    @staticmethod
    def initialize(vtfun, node1, k=10, nprobe=0):
        """Called during the initialziation of the virtual table function for a set of input
        parameters 'node1' (the source vector we are comparing against), 'k' (how many top-similar
        results to compute), and 'nprobe' (how deeply to search in the nearest-neighbor index).
        """
        # TO DO: the optional arguments don't work as expected, we always have to supply them it seems,
        #        otherwise the query breaks, figure out how to work around that...
        vtfun.input_vector = node1
        vtfun.input_k = k
        vtfun.input_nprobe = nprobe
        vtfun._result_rows = None

    @staticmethod
    def compute_result_rows(vtfun):
        """Called to compute a set of 'k' result rows for the set of input parameters stored by 'initialize'.
        Each row binds all output 'columns' specified above.  If no nearest-neighbor index is available
        for the associated vector store, this simply fails with an empty result.
        """
        vector = vtfun.input_vector
        k = vtfun.input_k
        nprobe = vtfun.input_nprobe
        vstore = vtfun.vector_store
        nnindex = vstore.get_nearest_neighbor_index() if vstore is not None else None
        if not nnindex or not nnindex.is_trained():
            return iter([])
        ndim = vstore.get_vector_ndim()
        dtype = vstore.get_vector_dtype()
        vector = np.frombuffer(vector, dtype=dtype)
        # TO DO: possibly increase 'k' here to account for different L2 vs. cosine distance ordering:
        D, V = nnindex.search(vector, k, nprobe=nprobe)
        result = []
        label = TopKCosineSimilarity.name
        vx = vector
        vxnorm = np.linalg.norm(vx)
        for row in vstore.get_vector_rows_by_id(V[0]):
            vy = row[4]
            # ['label', 'node2', 'vid', 'vrowid', 'vector', 'similarity']
            res = [label, row[1], row[0], row[3], vy.tobytes(), 0.0]
            vynorm = np.linalg.norm(vy)
            # make sure we don't return numpy floats:
            sim = float(np.dot(vx, vy) / (vxnorm * vynorm))
            res[-1] = sim
            result.append(res)
        result.sort(key=lambda x: x[-1], reverse=True)
        return iter(result)

    def get_code(self):
        """Create the virtual table code object and associate the appropriate vector store
        as determined by self.translate_call_to_sql().
        """
        if self.code is None:
            super().get_code()
            self.code.vector_store = self.xstore
        return self.code

    def translate_call_to_sql(self, query, clause, state):
        """Called by query.pattern_clause_to_sql() to translate a clause
        with a virtual graph pattern.  This primarily substitutes the
        appropriate virtual graph tables to use and computes the associated
        vector store.
        """
        node1 = clause[0]
        rel = clause[1]
        if rel.labels is None:
            return
        self.xstore, _ = self.get_argument_vector_store(query, node1.variable, state)
        # load here so we get the uniquified name registered with the connection:
        self.load()
        old_graph = node1._graph_table
        old_graph_alias = node1._graph_alias
        new_graph = self.get_name()
        new_graph_alias = old_graph_alias.replace(old_graph, new_graph)
        node1._graph_table = new_graph
        node1._graph_alias = new_graph_alias
        # TO DO: support this in query.py:
        #state.unregister_table_alias(old_graph, old_graph_alias)
        state.register_table_alias(new_graph, new_graph_alias)
        # prevent the generation of a label restriction based on the virtual graph name:
        rel.labels = None
        # now finish translation with standard translator:
        query.pattern_clause_to_sql(clause, new_graph_alias, state)

TopKCosineSimilarity('kvec_topk_cos_sim').define()
TopKCosineSimilarity('kvec_topk_cosine_similarity').define()

"""
# Example query:
> kgtk query --gc wikidata-20210215-dwd-v2-embeddings-test.sqlite3.db -i complexemb -i labels \
       --match 'comp: (x:Q40)-[]->(xv), \
                      (xv)-[r:kvec_topk_cos_sim {k: 20, nprobe: 8}]->(y), \
                lab:  (x)-[]->(xl), (y)-[]->(yl)' \
       --where 'x != y' \
       --return 'x, xl, y, yl, r.similarity' \
       --limit 10
node1	node2	node2	node2	similarity
Q40	'Austria'@en	Q28	'Hungary'@en	0.7745946049690247
Q40	'Austria'@en	Q794	'Iran'@en	0.7634432315826416
Q40	'Austria'@en	Q347	'Liechtenstein'@en	0.7329589128494263
Q40	'Austria'@en	Q155	'Brazil'@en	0.7318835854530334
Q40	'Austria'@en	Q79	'Egypt'@en	0.7268905639648438
Q40	'Austria'@en	Q35	'Denmark'@en	0.7170709371566772
Q40	'Austria'@en	Q212	'Ukraine'@en	0.7160522937774658
Q40	'Austria'@en	Q865	'Taiwan'@en	0.7146331071853638
Q40	'Austria'@en	Q928	'Philippines'@en	0.7101161479949951
Q40	'Austria'@en	Q298	'Chile'@en	0.7095608711242676
"""



"""Copy records from the first KGTK file to the output file,
adding ID values.
TODO: Need KgtkWriterOptions
"""
from argparse import Namespace, SUPPRESS

from kgtk.cli_argparse import KGTKArgumentParser, KGTKFiles


def parser():
    return {
        'help': 'Creating community detection from graph-tool using KGTK file',
        'description': 'Creating community detection from graph-tool ' +
        'using KGTK file, available options are blockmodel, nested and mcmc'
    }


def add_arguments_extended(parser: KGTKArgumentParser,
                           parsed_shared_args: Namespace):
    """
    Parse arguments
    Args:
        parser (argparse.ArgumentParser)
    """
    # import modules locally
    from kgtk.io.kgtkreader import KgtkReader, KgtkReaderOptions
    from kgtk.reshape.kgtkidbuilder import KgtkIdBuilder, KgtkIdBuilderOptions
    from kgtk.value.kgtkvalueoptions import KgtkValueOptions

    _expert: bool = parsed_shared_args._expert

    # This helper function makes it easy to suppress options from
    # The help message.  The options are still there, and initialize
    # what they need to initialize.
    def h(msg: str) -> str:
        if _expert:
            return msg
        else:
            return SUPPRESS

    parser.add_input_file(positional=True)
    parser.add_output_file()

    parser.add_argument('--method', dest='method', type=str,
                        default="blockmodel",
                        help="Specify the clustering method to use.")

    KgtkIdBuilderOptions.add_arguments(parser,
                                       expert=True)  # Show all the options.
    KgtkReader.add_debug_arguments(parser, expert=_expert)
    KgtkReaderOptions.add_arguments(parser, mode_options=True, expert=_expert)
    KgtkValueOptions.add_arguments(parser, expert=_expert)


def run(input_file: KGTKFiles,
        output_file: KGTKFiles,
        errors_to_stdout: bool = False,
        errors_to_stderr: bool = True,
        show_options: bool = False,
        verbose: bool = False,
        very_verbose: bool = False,
        method: str = "blockmodel",

        **kwargs  # Whatever KgtkFileOptions and KgtkValueOptions want.
        ) -> int:
    # import modules locally
    from pathlib import Path
    import sys
    import typing

    from graph_tool import Graph
    from graph_tool.inference.minimize import minimize_blockmodel_dl, \
        minimize_nested_blockmodel_dl
    import graph_tool

    from kgtk.io.kgtkreader import KgtkReader, KgtkReaderOptions
    from kgtk.io.kgtkwriter import KgtkWriter
    from kgtk.reshape.kgtkidbuilder import KgtkIdBuilder, KgtkIdBuilderOptions
    from kgtk.value.kgtkvalueoptions import KgtkValueOptions

    input_kgtk_file: Path = KGTKArgumentParser.get_input_file(input_file)
    output_kgtk_file: Path = KGTKArgumentParser.get_output_file(output_file)

    # Select where to send error messages, defaulting to stderr.
    error_file: typing.TextIO = sys.stdout if errors_to_stdout else sys.stderr

    # Build the option structures.
    idbuilder_options: KgtkIdBuilderOptions =\
    KgtkIdBuilderOptions.from_dict(kwargs)
    reader_options: KgtkReaderOptions = KgtkReaderOptions.from_dict(kwargs)
    value_options: KgtkValueOptions = KgtkValueOptions.from_dict(kwargs)

    # Show the final option structures for debugging and documentation.
    try:

        # First create the KgtkReader.  It provides parameters used by the ID
        # column builder. Next, create the ID column builder, which provides a
        # possibly revised list of column names for the KgtkWriter.  Create
        # the KgtkWriter.  Last, process the data stream.

        # Open the input file.
        kr: KgtkReader = KgtkReader.open(input_kgtk_file,
                                          error_file=error_file,
                                          options=reader_options,
                                          value_options=value_options,
                                          verbose=verbose,
                                          very_verbose=very_verbose,
        )

        g = Graph(directed=False)

        d = {}
        count = 0
        nodes = []
        edges = []
        for row in kr:
            if row[kr.node1_column_idx] not in d:
                d[row[kr.node1_column_idx]] = count
                count = count + 1
                nodes.append(row[kr.node1_column_idx])
            if row[kr.node2_column_idx] not in d:
                d[row[kr.node2_column_idx]] = count
                count = count + 1
                nodes.append(row[kr.node2_column_idx])
            edges.append((row[kr.node1_column_idx], row[kr.node2_column_idx]))

        vlist = g.add_vertex(len(d))

        for ele in edges:
            g.add_edge(g.vertex(d[ele[0]]), g.vertex(d[ele[1]]))

        if method == 'blockmodel':
            state = minimize_blockmodel_dl(g)
            arr = []

            for i in range(0, len(nodes)):
                arr.append('cluster_' + str(state.get_blocks()[i]))

            kw: KgtkWriter = KgtkWriter.open(["node1", "label", "node2"],
                                                 output_kgtk_file,
                                                 verbose=verbose,
                                                 very_verbose=very_verbose,
                                     )

            for i in range(0, len(nodes)):
                kw.write([nodes[i], 'in', arr[i]])

        elif method == 'nested':
            state = minimize_nested_blockmodel_dl(g)

            arr = []

            for i in range(0, len(nodes)):
                arr.append([str(i)])

            for i in range(0, len(state.levels)):
                if state.levels[i].get_B() == 1:
                    break
                for j in range(0, len(arr)):
                    arr[j].insert(0, str(state.levels[i].get_blocks()
                                         [arr[j][len(arr[j])-1]]))
            for i in range(0, len(nodes)):
                if len(arr[i]) > 0:
                    arr[i].pop()
                arr[i] = 'cluster_' + '_'.join(arr[i])

            kw: KgtkWriter = KgtkWriter.open(["node1", "label", "node2"],
                                                 output_kgtk_file,
                                                 verbose=verbose,
                                                 very_verbose=very_verbose,
                                     )
            for i in range(0, len(nodes)):
                kw.write([nodes[i], 'in', arr[i]])
        elif method == 'mcmc':
            state = minimize_blockmodel_dl(g)
            graph_tool.inference.mcmc.\
                mcmc_equilibrate(state, wait=1000, mcmc_args=dict(niter=10))

            dS, nattempts, nmoves = state.multiflip_mcmc_sweep(niter=1000)
            graph_tool.inference.mcmc.\
                mcmc_equilibrate(state, wait=10,
                                 nbreaks=2, mcmc_args=dict(niter=10))

            bs = []  # collect some partitions

            def collect_partitions(s):
                bs.append(s.b.a.copy())

            # Now we collect partitions for exactly 100,000 sweeps
            # of 10 sweeps:
            graph_tool.inference.mcmc.mcmc_equilibrate(
                state,
                force_niter=10000,
                mcmc_args=dict(niter=10),
                callback=collect_partitions)

            # Disambiguate partitions and obtain marginals
            pmode = graph_tool.inference.partition_modes.PartitionModeState(bs, converge=True)
            pv = list(pmode.get_marginal(g))
            m = list(pmode.get_max(g))

            kw: KgtkWriter =\
            KgtkWriter.open(["node1", "label", "node2", 'node2;prob'],
                                                 output_kgtk_file,
                                                 verbose=verbose,
                                                 very_verbose=very_verbose,
                                     )

            for i in range(0, len(nodes)):
                kw.write([nodes[i], 'in',
                         'cluster_' + str(m[i]), str(pv[i][m[i]]/sum(pv[i]))])

        kr.close()
        kw.close()
        return 0

    except SystemExit as e:
        raise KGTKException("Exit requested")
    except Exception as e:
        raise KGTKException(str(e))



import shutil
import tempfile
import unittest
from pathlib import Path

from kgtk.cli_entry import cli_entry
from kgtk.utils.convert_embeddings_format import ConvertEmbeddingsFormat
from kgtk.exceptions import KGTKException


class TestConvertEmbeddingsFormat(unittest.TestCase):
    def setUp(self) -> None:
        self.edge_file = 'data/convert_embeddings_edge.tsv'
        self.node_file = 'data/convert_embeddings_node.tsv'
        self.ground_truth_w2v = 'data/convert_embeddings_edge_w2v.txt'
        self.ground_truth_gprojector = 'data/convert_embeddings_edge_gprojector.tsv'
        self.ground_truth_gprojector_metadata = 'data/convert_embeddings_edge_gprojector_metadata.tsv'
        self.ground_truth_gprojector_metadata_selective = 'data/convert_embeddings_gprojector_metadata_selective.tsv'
        self.ground_truth_gprojector_metadata_edge = 'data/convert_embeddings_edge_gprojector_edge_metadata.tsv'

        self.temp_dir = tempfile.mkdtemp()

    def tearDown(self) -> None:
        shutil.rmtree(self.temp_dir)

    def test_convert_to_w2v(self):
        output = f'{self.temp_dir}/w2v.txt'
        cli_entry("kgtk", "--debug",
                  "convert-embeddings-format",
                  "-i",
                  self.edge_file,
                  "--input-property", "graph_embeddings",
                  "-o", output)

        f1 = open(self.ground_truth_w2v)
        f1_lines = set(f1.readlines())
        f1.close()
        with open(output) as f2:
            for line in f2:
                self.assertTrue(line in f1_lines)

    def test_convert_to_gprojector(self):
        output = f'{self.temp_dir}/gprojector.tsv'
        metadata_output = f'{self.temp_dir}/gprojector_metadata.tsv'
        cli_entry("kgtk", "--debug",
                  "convert-embeddings-format",
                  "-i",
                  self.edge_file,
                  "--output-format", "gprojector",
                  "--node-file", self.node_file,
                  "--input-property", "graph_embeddings",
                  "--metadata-file", metadata_output,
                  "-o", output)

        f1 = open(self.ground_truth_gprojector)
        f1_lines = set(f1.readlines())
        f1.close()
        with open(output) as f2:
            for line in f2:
                self.assertTrue(line in f1_lines)

        f1_metadata = open(self.ground_truth_gprojector_metadata)
        f1_metadata_lines = f1_metadata.readlines()
        f1_metadata.close()
        with open(metadata_output) as f2:
            for line in f2:
                self.assertTrue(line in f1_metadata_lines)

    def test_convert_to_gprojector_selective_metadata(self):
        output = f'{self.temp_dir}/gprojector.tsv'
        metadata_output = f'{self.temp_dir}/gprojector_metadata_selective.tsv'
        cli_entry("kgtk", "--debug",
                  "convert-embeddings-format",
                  "-i",
                  self.edge_file,
                  "--output-format", "gprojector",
                  "--node-file", self.node_file,
                  "--metadata-columns", "label,type_label",
                  "--input-property", "graph_embeddings",
                  "--metadata-file", metadata_output,
                  "-o", output)

        f1 = open(self.ground_truth_gprojector)
        f1_lines = set(f1.readlines())
        f1.close()
        with open(output) as f2:
            for line in f2:
                self.assertTrue(line in f1_lines)

        f1_metadata = open(self.ground_truth_gprojector_metadata_selective)
        f1_metadata_lines = f1_metadata.readlines()
        f1_metadata.close()
        with open(metadata_output) as f2:
            for line in f2:
                self.assertTrue(line in f1_metadata_lines)

    def test_convert_to_gprojector_metadata_edge_file(self):
        output = f'{self.temp_dir}/gprojector.tsv'
        metadata_output = f'{self.temp_dir}/gprojector_metadata.tsv'
        cli_entry("kgtk", "--debug",
                  "convert-embeddings-format",
                  "-i",
                  self.edge_file,
                  "--output-format", "gprojector",
                  "--metadata-columns", "node1_label,type",
                  "--input-property", "graph_embeddings",
                  "--metadata-file", metadata_output,
                  "-o", output)

        f1 = open(self.ground_truth_gprojector)
        f1_lines = set(f1.readlines())
        f1.close()
        with open(output) as f2:
            for line in f2:
                self.assertTrue(line in f1_lines)

        f1_metadata = open(self.ground_truth_gprojector_metadata_edge)
        f1_metadata_lines = f1_metadata.readlines()
        f1_metadata.close()
        with open(metadata_output) as f2:
            for line in f2:
                self.assertTrue(line in f1_metadata_lines)

    def test_convert_to_gprojector_columns_not_in_node_file(self):
        output = f'{self.temp_dir}/gprojector.tsv'
        metadata_output = f'{self.temp_dir}/gprojector_metadata.tsv'

        cef: ConvertEmbeddingsFormat = ConvertEmbeddingsFormat(input_file=Path(self.edge_file),
                                                               output_file=Path(output),
                                                               node_file=Path(self.node_file),
                                                               output_format='gprojector',
                                                               input_property='graph_embeddings',
                                                               output_metadata_file=Path(metadata_output),
                                                               metadata_columns="blaaah"
                                                               )

        with self.assertRaises(KGTKException):
            cef.process()

    def test_convert_to_gprojector_columns_not_in_edge_file(self):
        output = f'{self.temp_dir}/gprojector.tsv'
        metadata_output = f'{self.temp_dir}/gprojector_metadata.tsv'

        cef: ConvertEmbeddingsFormat = ConvertEmbeddingsFormat(input_file=Path(self.edge_file),
                                                               output_file=Path(output),
                                                               output_format='gprojector',
                                                               input_property='graph_embeddings',
                                                               output_metadata_file=Path(metadata_output),
                                                               metadata_columns="blaaah"
                                                               )

        with self.assertRaises(KGTKException):
            cef.process()



import shutil
import unittest
import tempfile
import pandas as pd
from kgtk.cli_entry import cli_entry
from kgtk.exceptions import KGTKException

class TestKGTKImportConceptPairs(unittest.TestCase):
    def setUp(self) -> None:
        self.temp_dir=tempfile.mkdtemp()

    def tearDown(self) -> None:
        shutil.rmtree(self.temp_dir)

    def test_kgtk_import_concept_pairs(self):
        cli_entry("kgtk", "import-concept-pairs", "-i", "data/synonyms.txt", "--source", "RG", "--relation", "/r/Synonym", "-o", f'{self.temp_dir}/roget_syn.tsv')

        df = pd.read_csv(f'{self.temp_dir}/roget_syn.tsv', sep='\t')

        self.assertEqual(len(df.columns), 9)

        for i, row in df.iterrows():
            self.assertTrue(row['relation']=='/r/Synonym')
        print('ROGET', df)



import shutil
import unittest
import tempfile
import pandas as pd
from kgtk.cli_entry import cli_entry
from kgtk.exceptions import KGTKException

from graph_tool.all import *

class TestExportGT(unittest.TestCase):
    def setUp(self) -> None:
        self.temp_dir=tempfile.mkdtemp()

    def tearDown(self) -> None:
        shutil.rmtree(self.temp_dir)

    def test_export_gt(self):
        cli_entry("kgtk", "export-gt", "-i", "data/sample_kgtk_edge_file.tsv", "--directed", "--log", f'{self.temp_dir}/log.txt', "-o", f'{self.temp_dir}/graph.gt')

        

        g=load_graph(f'{self.temp_dir}/graph.gt')

        self.assertEqual(g.num_edges(), 287)
        self.assertEqual(g.num_vertices(), 287)




import os
import json
import subprocess
from pathlib import Path
from typing import List
from kgtk.files_config import files_config
import shutil

always_print_env_variables = {
    'EXAMPLES_DIR',
    'USE_CASES_DIR',
    'GRAPH',
    'OUT',
    'TEMP',
    'STORE',
    'kgtk',
    'kypher',
    'KGTK_GRAPH_CACHE',
    'KGTK_OPTION_DEBUG',
    'KGTK_LABEL_FILE'
}


class ConfigureKGTK(object):
    def __init__(self, file_list: List[str], kgtk_path: str = None, input_files_url: str = None):

        self.graph_files = files_config
        self.files = file_list
        self.INPUT_FILES_URL = "https://github.com/usc-isi-i2/kgtk-tutorial-files/raw/main/datasets/arnold" \
            if input_files_url is None else input_files_url

        self.kgtk_environment_variables = set()
        self.user_home = str(Path.home())
        print(f'User home: {self.user_home}')

        self.default_folder = "isi-kgtk-tutorial"

        self.current_dir = os.getcwd()
        print(f'Current dir: {self.current_dir}')
        if kgtk_path is not None:
            self.kgtk_path = kgtk_path
        else:
            self.kgtk_path = Path(self.current_dir).parent.absolute()
        print(f'KGTK dir: {self.kgtk_path}')
        self.examples_dir = f"{self.kgtk_path}/examples"
        self.use_cases_dir = f"{self.kgtk_path}/use-cases"
        print(f'Use-cases dir: {self.use_cases_dir}')

        os.environ['EXAMPLES_DIR'] = self.examples_dir
        os.environ['USE_CASES_DIR'] = self.use_cases_dir

        os.environ['KGTK_OPTION_DEBUG'] = "false"

        self.kgtk_environment_variables.add('EXAMPLES_DIR')
        self.kgtk_environment_variables.add('USE_CASES_DIR')
        self.kgtk_environment_variables.add('KGTK_OPTION_DEBUG')

    def configure_kgtk(self,
                       input_graph_path: str = None,
                       output_path: str = None,
                       project_name: str = "kgtk",
                       graph_cache_path: str = None,
                       json_config_file: str = None,
                       additional_files: dict = None,
                       debug=False):
        """
        configures the environment for a jupyter notebook.
        :param debug: add --debug flag in the kypher command
        :param additional_files: additional files in dict format to be used
        :param input_graph_path: path to the input graph files. By default it'll create a folder "isi-kgtk-tutorial" in
        user home and download files from github
        :param output_path: path where the output and temp files will be created. By default, "isi-kgtk-tutorial-out" in
        user home
        :param project_name: output folder name, 'kgtk' by default
        :param graph_cache_path: absolute path to the '.db' file. If not specified, create a new cache file in output
        path
        :param json_config_file: absolute path to json file which contains "file_key": "file_name" key-value pairs. By
        default, read a file from kgtk repo
        :return:
        """

        if json_config_file is not None:
            try:
                _files_config = json.load(open(json_config_file))
                self.graph_files.update(_files_config)
                self.files.extend(list(_files_config))
            except Exception as e:
                print(e)

        if additional_files is not None:
            self.graph_files.update(additional_files)
            self.files.extend(list(additional_files))

        # If the input graph path is not None, it is assumed it has the files required
        if input_graph_path is None:
            input_graph_path = f"{self.user_home}/{self.default_folder}/{project_name}_input"

            # if the input folder already exists from a previous run, delete it
            if Path(input_graph_path).exists():
                shutil.rmtree(input_graph_path)

            Path(input_graph_path).mkdir(parents=True, exist_ok=True)
            self.download_tutorial_files(input_graph_path)

        os.environ['GRAPH'] = input_graph_path
        self.kgtk_environment_variables.add('GRAPH')

        for key in self.graph_files:
            os.environ[key] = f"{input_graph_path}/{self.graph_files[key]}"
            self.kgtk_environment_variables.add(key)

        if output_path is None:
            output_project_path = f"{self.user_home}/{self.default_folder}/{project_name}"
        else:
            output_project_path = f"{output_path}/{project_name}"

        temp_path = f"{output_project_path}/temp.{project_name}"

        Path(output_project_path).mkdir(parents=True, exist_ok=True)
        Path(temp_path).mkdir(parents=True, exist_ok=True)

        os.environ['OUT'] = output_project_path
        os.environ['TEMP'] = temp_path
        self.kgtk_environment_variables.add('OUT')
        self.kgtk_environment_variables.add('TEMP')

        if graph_cache_path is None:
            graph_cache_path = f"{temp_path}/wikidata.sqlite3.db"
        os.environ['STORE'] = graph_cache_path
        self.kgtk_environment_variables.add('STORE')

        os.environ['KGTK_GRAPH_CACHE'] = os.environ['STORE']
        self.kgtk_environment_variables.add('KGTK_GRAPH_CACHE')

        kgtk = "kgtk --debug" if debug else "kgtk"
        os.environ['kgtk'] = kgtk
        self.kgtk_environment_variables.add('kgtk')

        kypher = kgtk + " query --graph-cache " + os.environ['STORE']
        os.environ['kypher'] = kypher
        self.kgtk_environment_variables.add('kypher')

    def download_tutorial_files(self, graph_path: str):
        if not graph_path.endswith('/'):
            graph_path += '/'

        for f in self.files:
            if f in self.graph_files:
                url = f"{self.INPUT_FILES_URL}/{self.graph_files[f]}"
                cmd = f" wget {url} --directory-prefix={graph_path}"
                print(subprocess.getoutput(cmd))
            else:
                print(f"File: {f} not present in files config: {self.graph_files}")

    def print_env_variables(self):
        for key in self.kgtk_environment_variables:
            if key in always_print_env_variables:
                print(f"{key}: {os.environ[key]}")

        for key in self.files:
            print(f"{key}: {os.environ[key]}")

    def load_files_into_cache(self, files: List[str] = None):
        """
        Loads files into graph cache. The keys in this list should be in json_config_file
        :return:
        """
        kypher_command = f"{os.environ['kypher']}"

        if files is None:
            files = self.files

        for f_key in files:
            kypher_command += f" -i \"{os.environ[f_key]}\" --as {f_key} "
        kypher_command += " --limit 3"
        print(kypher_command)
        print(subprocess.getoutput(kypher_command))

    def print_kgtk_file_names(self, file_names: List[str] = None):
        if file_names is not None:
            if not type(file_names) == list:
                file_names = [file_names]
            for file_name in file_names:
                if file_name not in self.graph_files:
                    print(f"Unknown file name: {file_name}.")
                else:
                    print(f"{file_name}: {self.graph_files[file_name]}")
        else:
            print("Printing a list of all known KGTK file names...")
            for key in self.graph_files:
                print(f"{key}: {self.graph_files[key]}")



"""This randomly samples a KGTK file, dividing it into an output file and an optional reject file.

A probability option, `--probability frac`, determines the probability that
an input record is passed to the standard output file. The probability ranges
from 0.0 to 1.0, with 1 being the default.  The number of output records
may not match the number of input records times the probability.

The probability value must not be negative, and it must not be greater than 1.

Alternatively, `--input-size N' and '--sample-size n' may be provided.  The
sampling probability will be computed as n/N. The number of output records may not
exactly match the desired count unless `--exact` is specified.  `--exact`
consumes more memory on large input files.

The input count, if specified, must be positive.  The desired count, if specified,
must be positive.

Finally, `--sample-size n` may be provided without `--input-size N`.  Exactly
`n` records will be selected, unless the input file has fewer than `n` records.
The selected records will be buffered in memory as the input file is processed,
so a significant amount of memory may be needed if `n` is large.

TODO: Optionally raise an exception if the sample size is greater than the
input size (prespecified or the actual number of input records, as applicable).

--mode=NONE is default.

TODO: Need KgtkWriterOptions

"""

from argparse import Namespace, SUPPRESS
import typing

from kgtk.cli_argparse import KGTKArgumentParser, KGTKFiles

def parser():
    return {
        'help': 'Randomly sample a KGTK file.',
        'description': 'This utility randomly samples a KGTK file, dividing it into an optput file and an optional reject file. ' +
        'The probability of an input record being passed to the output file is controlled by `--probability n`, where ' +
        '`n` ranges from 0 to 1. ' +
        '\n\nThis command defaults to --mode=NONE so it will work with TSV files that do not follow KGTK column naming conventions.' +
        '\n\nAdditional options are shown in expert help.\nkgtk --expert sample --help'
    }


def add_arguments_extended(parser: KGTKArgumentParser, parsed_shared_args: Namespace):
    """
    Parse arguments
    Args:
        parser (argparse.ArgumentParser)
    """
    from kgtk.io.kgtkreader import KgtkReader, KgtkReaderOptions, KgtkReaderMode
    from kgtk.io.kgtkwriter import KgtkWriter
    from kgtk.utils.argparsehelpers import optional_bool
    from kgtk.value.kgtkvalueoptions import KgtkValueOptions

    _expert: bool = parsed_shared_args._expert

    # This helper function makes it easy to suppress options from
    # The help message.  The options are still there, and initialize
    # what they need to initialize.
    def h(msg: str)->str:
        if _expert:
            return msg
        else:
            return SUPPRESS

    parser.add_input_file()
    parser.add_output_file()
    parser.add_output_file(who="The KGTK reject file for records that fail the filter.",
                           dest="reject_file",
                           options=["--reject-file"],
                           metavar="REJECT_FILE",
                           optional=True)

    parser.add_argument("--probability", dest="probability", type=float, default=1.0,
                        help="The probability of passing an input record to the output file (default=%(default)d).")
                        
    parser.add_argument("--seed", dest="seed", type=int,
                        help="The optional random number generator seed (default=None).")

    parser.add_argument("--input-size", dest="input_size", type=int,
                        help="The optional number of input records (default=None).")

    parser.add_argument("--sample-size", dest="sample_size", type=int,
                        help="The optional desired number of output records (default=None).")

    parser.add_argument("--exact", dest="exact", metavar="True|False",
                        help="Ensure that exactly the desired sample size is extracted when " +
                        "--input-size and --sample-size are supplied. (default=%(default)s).",
                        type=optional_bool, nargs='?', const=True, default=False)

    parser.add_argument(      "--output-format", dest="output_format", help=h("The file format (default=kgtk)"), type=str,
                              choices=KgtkWriter.OUTPUT_FORMAT_CHOICES)

    KgtkReader.add_debug_arguments(parser, expert=_expert)
    KgtkReaderOptions.add_arguments(parser,
                                    mode_options=True,
                                    default_mode=KgtkReaderMode.NONE,
                                    expert=_expert)
    KgtkValueOptions.add_arguments(parser, expert=_expert)

def run(input_file: KGTKFiles,
        output_file: KGTKFiles,
        reject_file: KGTKFiles,

        probability: float,
        seed: typing.Optional[int],
        input_size: typing.Optional[int],
        sample_size: typing.Optional[int],
        exact: bool,
        output_format: str,

        errors_to_stdout: bool = False,
        errors_to_stderr: bool = True,
        show_options: bool = False,
        verbose: bool = False,
        very_verbose: bool = False,

        **kwargs # Whatever KgtkFileOptions and KgtkValueOptions want.
)->int:
    # import modules locally
    from collections import deque
    from pathlib import Path
    from queue import PriorityQueue
    import random
    import sys
    import typing
    
    from kgtk.exceptions import KGTKException
    from kgtk.io.kgtkreader import KgtkReader, KgtkReaderOptions, KgtkReaderMode
    from kgtk.io.kgtkwriter import KgtkWriter
    from kgtk.value.kgtkvalueoptions import KgtkValueOptions

    input_file_path: Path = KGTKArgumentParser.get_input_file(input_file)
    output_file_path: Path = KGTKArgumentParser.get_output_file(output_file)
    reject_file_path: typing.Optional[Path] = KGTKArgumentParser.get_optional_output_file(reject_file, who="KGTK reject file")


    # Select where to send error messages, defaulting to stderr.
    error_file: typing.TextIO = sys.stdout if errors_to_stdout else sys.stderr

    # TODO: check that at most one input file is stdin?

    # Build the option structures.
    reader_options: KgtkReaderOptions = KgtkReaderOptions.from_dict(kwargs, mode=KgtkReaderMode.NONE)
    value_options: KgtkValueOptions = KgtkValueOptions.from_dict(kwargs)

    # Show the final option structures for debugging and documentation.
    if show_options:
        print("--input-file=%s" % str(input_file_path), file=error_file, flush=True)
        print("--output-file=%s" % str(output_file_path), file=error_file, flush=True)
        if reject_file_path is not None:
            print("--reject-file=%s" % str(reject_file_path), file=error_file)
        print("--probability=%s" % str(probability), file=error_file, flush=True)
        if input_size is not None:
            print("--input-size=%d" % input_size, file=error_file, flush=True)
        if sample_size is not None:
            print("--sample-size=%d" % sample_size, file=error_file, flush=True)
        print("--exact=%s" % str(exact), file=error_file, flush=True)
        if seed is not None:
            print("--seed=%d" % seed, file=error_file, flush=True)
        reader_options.show(out=error_file)
        value_options.show(out=error_file)
        print("=======", file=error_file, flush=True)

    if probability < 0.0:
        raise KGTKException("The probability (%f) must not be negative." % probability)

    if probability > 1.0:
        raise KGTKException("The probability (%f) must not be greater than 1.0." % probability)

    if input_size is not None and input_size <= 0:
        raise KGTKException("The input count (%d) must be positive." % input_size)

    if sample_size is not None and sample_size <= 0:
        raise KGTKException("The desired count (%d) must be positive." % input_size)


    # Use our own random number generator.  If the seed is not supplied, a
    # system-provided random source will be used, or the clock time as a
    # fallback.
    rg: random.Random = random.Random(seed)

    sample_set: typing.Optional[typing.Set[int]] = None
    priorityq: typing.Optional[PriorityQueue] = None  # TODO: can we provide a more complete type hint?
    if sample_size is not None:
        if input_size is None:
            priorityq = PriorityQueue()
        else:
            if sample_size > input_size:
                probability = 1.0
            else:
                if exact:
                    sample_set = set(rg.sample(range(input_size), sample_size))
                else:
                    probability = sample_size / input_size

    try:
        kr: KgtkReader = KgtkReader.open(input_file_path,
                                         options=reader_options,
                                         value_options = value_options,
                                         error_file=error_file,
                                         verbose=verbose,
                                         very_verbose=very_verbose,
                                         )

        output_mode: KgtkWriter.Mode = KgtkWriter.Mode.NONE
        if kr.is_edge_file:
            output_mode = KgtkWriter.Mode.EDGE
            if verbose:
                print("Opening the output edge file: %s" % str(output_file_path), file=error_file, flush=True)

        elif kr.is_node_file:
            output_mode = KgtkWriter.Mode.NODE
            if verbose:
               print("Opening the output node file: %s" % str(output_file_path), file=error_file, flush=True)

        else:
            if verbose:
                print("Opening the output file: %s" % str(output_file_path), file=error_file, flush=True)

        kw: KgtkWriter = KgtkWriter.open(kr.column_names,
                                         output_file_path,
                                         use_mgzip=reader_options.use_mgzip, # Hack!
                                         mgzip_threads=reader_options.mgzip_threads, # Hack!
                                         gzip_in_parallel=False,
                                         mode=output_mode,
                                         output_format=output_format,
                                         error_file=error_file,
                                         verbose=verbose,
                                         very_verbose=very_verbose)

        rkw: typing.Optional[KgtkWriter] = None
        if reject_file_path is not None:
            rkw = KgtkWriter.open(kr.column_names,
                                  reject_file_path,
                                  use_mgzip=reader_options.use_mgzip, # Hack!
                                  mgzip_threads=reader_options.mgzip_threads, # Hack!
                                  gzip_in_parallel=False,
                                  mode=output_mode,
                                  output_format=output_format,
                                  error_file=error_file,
                                  verbose=verbose,
                                  very_verbose=very_verbose)

        def copy_sample_set() -> typing.Tuple[int, int, int]:
            input_count: int = 0
            output_count: int = 0
            reject_count: int = 0

            row: typing.List[str]
            for row in kr:
                if input_count in sample_set:
                    kw.write(row)
                    output_count += 1

                elif rkw is not None:
                    rkw.write(row)
                    reject_count += 1

                input_count += 1
            return input_count, output_count, reject_count

        def copy_probably() -> typing.Tuple[int, int, int]:
            input_count: int = 0
            output_count: int = 0
            reject_count: int = 0

            row: typing.List[str]
            for row in kr:
                if probability > rg.random():
                    kw.write(row)
                    output_count += 1

                elif rkw is not None:
                    rkw.write(row)
                    reject_count += 1

                input_count += 1
            return input_count, output_count, reject_count

        def fill_priority_queue() -> typing.Tuple[int, int, int]:
            input_count: int = 0
            queued_count: int = 0
            reject_count: int = 0

            row: typing.List[str]
            for row in kr:
                priority = rg.random()
                priorityq.put((priority, row))
                queued_count += 1

                if queued_count > sample_size:
                    priority, row = priorityq.get()
                    queued_count -= 1
                    if rkw is not None:
                        rkw.write(row)
                        reject_count += 1

                input_count += 1
            return input_count, queued_count, reject_count

        def write_priority_queue(queued_count: int):

            priority: int
            row: typing.List[str]

            while queued_count > 0:
                queued_count -= 1
                priority, row = priorityq.get()
                kw.write(row)

        input_count: int
        output_count: int
        reject_count: int
        if priorityq is not None:
            input_count, output_count, reject_count = fill_priority_queue()
            write_priority_queue(output_count)
        else:
            input_count, output_count, reject_count = copy_sample_set() if sample_set is not None else copy_probably()

        kr.close()
        kw.close()
        if rkw is not None:
            rkw.close()
            
        if verbose:
            print("Read %d records, wrote %d records and %d reject records." % (input_count, output_count, reject_count),
                  file=error_file, flush=True)

    except SystemExit as e:
        raise KGTKException("Exit requested")
    except Exception as e:
        raise KGTKException(str(e))



from rdflib.plugins.stores.sparqlstore import SPARQLUpdateStore
from rdflib.query import ResultException


class TruthyUpdater:
    def __init__(self, endpoint, dryrun=False, user=None, passwd=None):
        self.endpoint = SPARQLUpdateStore(endpoint,
                                          default_query_method='POST')
        if user:
            self.endpoint.setCredentials(user, passwd)
        self.dryrun = dryrun

    def build_truthy(self, np_list):
        self.insert_truthy_rank(np_list)
        self.delete_normal_rank(np_list)

    def update(self, query):
        if self.dryrun:
            action = 'INSERT' if 'INSERT' in query else 'DELETE'
            query = query.replace(action, 'CONSTRUCT')
            # print(query)
            try:
                res = self.endpoint.query(query)
                print('### About to {} following triples:'.format(action))
                for row in res:
                    print(' '.join(e.n3() for e in row))
            except ResultException:
                pass
        else:
            self.endpoint.update(query)

    def insert_truthy_rank(self, np_list):
        values = ' '.join('( wd:%(node)s p:%(p)s ps:%(p)s psn:%(p)s wdt:%(p)s wdtn:%(p)s )' % {'node': n, 'p': p}
                          for n, p in np_list)
        query = '''
        INSERT {
          ?statement a wikibase:BestRank .
          ?node ?wdt ?psv ;
                ?wdtn ?psnv .
        } WHERE {
          %s
            {
                ?node ?p ?statement .
                ?statement wikibase:rank wikibase:PreferredRank ;
                           ?ps ?psv .
                OPTIONAL { ?statement ?psn ?psnv }
                FILTER NOT EXISTS { ?statement a wikibase:BestRank }
            }
            UNION
            {
              ?node ?p ?statement .
              ?statement wikibase:rank wikibase:NormalRank ;
                         ?ps ?psv .
              OPTIONAL { ?statement ?psn ?psnv }
              FILTER NOT EXISTS { ?statement a wikibase:BestRank }
              FILTER NOT EXISTS { ?node ?p [ wikibase:rank wikibase:PreferredRank ] }
            }
        }
        ''' % ('VALUES (?node ?p ?ps ?psn ?wdt ?wdtn ) { %s }' % values)
        self.update(query)

    def delete_normal_rank(self, np_list):
        values = ' '.join('( wd:%(node)s p:%(p)s wdt:%(p)s wdtn:%(p)s )' % {'node': n, 'p': p} for n, p in np_list)
        query = '''
          DELETE {
            ?statement a wikibase:BestRank .
            ?node ?wdt ?value ;
              ?wdtn ?no .
          } WHERE {
            %s
            ?node ?p ?statement ;
                  ?p [ wikibase:rank wikibase:PreferredRank ] .
            ?statement a wikibase:BestRank ;
                       wikibase:rank wikibase:NormalRank .
          }
        ''' % ('VALUES (?node ?p ?wdt ?wdtn ) { %s }' % values)
        self.update(query)



from typing import List, Union
from kgtk.knowledge_graph.ontology import Ontology
from kgtk.knowledge_graph.subject import Subject
from kgtk.knowledge_graph.node import URI, Literal
from kgtk.knowledge_graph.shacl import SHACL
from kgtk.utils import deprecated
import json


class KGSchema(object):
    """
    This class define the schema for a knowledge graph object.
    Create a knowledge graph schema according to the master config the user defined in myDIG UI
    """

    def __init__(self, content=None):
        self.ontology = Ontology()
        self.shacl = SHACL()
        self.need_convert = False
        self.legacy = False
        if content:
            self.legacy = True
            self.need_convert = True
            self.add_schema(content, 'master_config')

    def add_schema(self, content: str, format: str):
        """
        Add schema file into the ontology.
        :param content: schema content
        :param format: schema format, can be in 'master_config' or any RDF format
        """
        self.need_convert = True
        if format == 'master_config':
            if isinstance(content, dict):
                config = content
            else:
                config = json.loads(content)
            self._add_master_config(config)
        else:
            self.ontology.parse(content, format)

    def add_shacl(self, content: str, format: str):
        """
        Add SHACL file into the SHACL
        """
        self.shacl.parse(content, format)

    def _merge_ontology(self):
        self.shacl.add_ontology(self.ontology)
        self.need_convert = False

    def validate(self, data_graph, inference=None):
        if self.need_convert:
            self._merge_ontology()
        if inference:
            conforms, results_graph = self.shacl.validate(data_graph, self.ontology, inference)
        else:
            conforms, results_graph = self.shacl.validate(data_graph)
        return conforms, results_graph

    def _add_master_config(self, config):
        self.ontology._ns.bind_for_master_config()
        try:
            for field in config["fields"]:
                t = Subject(URI(field))
                if config["fields"][field]["type"] == "kg_id":
                    t.add_property(URI('rdf:type'), URI('owl:ObjectProperty'))
                elif config["fields"][field]["type"] == "number":
                    t.add_property(URI('rdf:type'), URI('owl:DatatypeProperty'))
                elif config["fields"][field]["type"] == "date":
                    t.add_property(URI('rdf:type'), URI('owl:DatatypeProperty'))
                elif config["fields"][field]["type"] == "location":
                    t.add_property(URI('rdf:type'), URI('owl:DatatypeProperty'))
                    t.add_property(URI('rdf:range'), URI('xsd:string'))
                else:
                    t.add_property(URI('rdf:type'), URI('owl:DatatypeProperty'))
                    t.add_property(URI('rdf:range'), URI('xsd:string'))
                if "description" in config["fields"][field] and config["fields"][field]["description"]:
                    t.add_property(URI('rdf:comment'), Literal(config["fields"][field]["description"]))
                self.ontology.add_subject(t)
        except KeyError as key:
            print(str(key) + " not in config")

    def is_valid(self, s_types: List[URI], p: URI, o_types: List[URI]) -> bool:
        """
        Check if it's a valid triple by checking the property's domain and range
        :param s_types: the types of the subject
        :param p: the property
        :param o_types: the types of the object
        :return: bool
        """
        return self.ontology.is_valid(s_types, p, o_types)

    @property
    def fields(self) -> List[str]:
        """
        Return a list of all fields that are defined in master config
        """
        return [self.ontology._ns.qname(uri) for uri in
                self.ontology.object_properties | self.ontology.datatype_properties]

    @deprecated()
    def has_field(self, field_name: str) -> bool:
        """
        Return true if the schema has the field, otherwise false
        """
        property_ = self.ontology._resolve_uri(URI(field_name))
        return property_ in self.ontology.object_properties or property_ in self.ontology.datatype_properties

    @deprecated()
    def parse_field(self, field_name: str):
        """
        Return the property URI of the field
        """
        property_ = self.ontology._resolve_uri(URI(field_name))
        return property_

    @deprecated()
    def field_type(self, field_name: str, value: object) -> Union[Subject, URI, Literal, None]:
        """
        Return the type of a field defined in schema, if field not defined, return None
        """
        property_ = self.ontology._resolve_uri(URI(field_name))
        if property_ in self.ontology.object_properties:
            return value if isinstance(value, Subject) else URI(value)
        elif property_ in self.ontology.datatype_properties:
            # TODO: check Literal type
            return Literal(value)
        else:
            return None



import graph_tool as gtmain
import graph_tool.all as gtall
import numpy as np  # type: ignore
from collections import defaultdict

import matplotlib.pyplot as plt  # type: ignore

plt.rcParams.update({'font.size': 12})

import seaborn as sns  # type: ignore

sns.set_style("whitegrid")


#### BASIC STATS ####

def get_num_nodes(g):
    return g.num_vertices()


def get_num_edges(g):
    return g.num_edges()


#### DEGREES ####

def compute_avg_node_degree(g, direction):
    return gtmain.stats.vertex_average(g, direction)


def compute_node_degree_hist(g, direction):
    return gtall.vertex_hist(g, direction, float_count=False)


def get_degree_maxn_counts(g, direction):
    return list(compute_node_degree_hist(g, direction)[0])[:10]


def plot_degrees(degrees, plottype='loglog', base=10, xlabel='', ylabel='', title=''):
    plt.loglog(degrees, basex=base, basey=base)
    plt.ylabel(ylabel)
    plt.xlabel(xlabel)
    plt.title(title)
    plt.show()


#### CENTRALITY ####

def compute_betweenness(g):
    bn, be = gtmain.centrality.betweenness(g)
    return bn, be


def compute_pagerank(g):
    v_pr = g.new_vertex_property('float')
    gtmain.centrality.pagerank(g, prop=v_pr)
    return v_pr


def compute_hits(g):
    hits_eig, v_hubs, v_auth = gtmain.centrality.hits(g)
    return hits_eig, v_hubs, v_auth


def get_max_node(g, prop):
    max_pr = 0.0
    max_pr_vertex = None
    for v in g.vertices():
        vertex_pr = g.vp[prop][v]
        if vertex_pr > max_pr:
            max_pr = vertex_pr
            max_pr_vertex = g.vp['_graphml_vertex_id'][v]

    return max_pr, max_pr_vertex


def get_topn_indices(g, prop, n, print_prop):
    a = g.vp[prop].a
    ind = np.argpartition(a, -n)[-n:]
    result = []
    for i in ind:
        result.append([i, g.vp[print_prop][i], g.vp[prop][i]])
    return sorted(result, key=lambda x: x[-1], reverse=True)


#### RUN ALL STATS ####

def compute_stats(g, direction):
    avg_degree, stdev_degree = compute_avg_node_degree(g, direction)
    return {
        'num_nodes': get_num_nodes(g),
        'num_edges': get_num_edges(g),
        'avg_degree': avg_degree,
        'degree_maxn_counts': get_degree_maxn_counts(g, direction),
        'stdev_degree': stdev_degree
    }


def get_topN_relations(g, N=10, pred_property='predicate'):
    rel_freq = defaultdict(int)
    for i, e in enumerate(g.edges()):
        r = g.edge_properties[pred_property][e]
        rel_freq[r] += 1
    return sorted(rel_freq.items(), key=lambda x: x[1], reverse=True)[:N]



"""Copy records from the first KGTK file to the output file,
exploding a column into several data type-specific columns.

"""

from argparse import ArgumentParser, Namespace
import attr
from pathlib import Path
import sys
import typing

from kgtk.kgtkformat import KgtkFormat
from kgtk.io.kgtkreader import KgtkReader, KgtkReaderOptions
from kgtk.io.kgtkwriter import KgtkWriter
from kgtk.utils.argparsehelpers import optional_bool
from kgtk.value.kgtkvalue import KgtkValue, KgtkValueFields
from kgtk.value.kgtkvalueoptions import KgtkValueOptions

@attr.s(slots=True, frozen=True)
class KgtkExplode(KgtkFormat):
    input_file_path: typing.Optional[Path] = attr.ib(validator=attr.validators.optional(attr.validators.instance_of(Path)))

    output_file_path: typing.Optional[Path] = attr.ib(validator=attr.validators.optional(attr.validators.instance_of(Path)))

    reader_options: KgtkReaderOptions= attr.ib(validator=attr.validators.instance_of(KgtkReaderOptions))
    value_options: KgtkValueOptions= attr.ib(validator=attr.validators.instance_of(KgtkValueOptions))

    output_format: typing.Optional[str] = attr.ib(validator=attr.validators.optional(attr.validators.instance_of(str)))

    column_name: str = attr.ib(validator=attr.validators.instance_of(str), default=KgtkFormat.NODE2)

    prefix: str = attr.ib(validator=attr.validators.instance_of(str), default= KgtkFormat.NODE2 + ";" + KgtkFormat.KGTK_NAMESPACE)
                               
    field_names: typing.Optional[typing.List[str]] = \
        attr.ib(validator=attr.validators.optional(attr.validators.deep_iterable(member_validator=attr.validators.instance_of(str),
                                                                                 iterable_validator=attr.validators.instance_of(list))),
                default=None)

    type_names: typing.Optional[typing.List[str]] = \
        attr.ib(validator=attr.validators.optional(attr.validators.deep_iterable(member_validator=attr.validators.instance_of(str),
                                                                                 iterable_validator=attr.validators.instance_of(list))),
                default=None)

    overwrite_columns: bool = attr.ib(validator=attr.validators.instance_of(bool), default=False)
    expand_list: bool = attr.ib(validator=attr.validators.instance_of(bool), default=False)
                               
    # TODO: find working validators to allow this parameter to be optional.
    # value_options: KgtkValueOptions = attr.ib(default=None,
    #                                           converter=attr.converters.default_if_none(factory=KgtkValueOptions.default),
    #                                           validator=attr.validators.instance_of(KgtkValueOptions))

    error_file: typing.TextIO = attr.ib(default=sys.stderr)
    verbose: bool = attr.ib(validator=attr.validators.instance_of(bool), default=False)
    very_verbose: bool = attr.ib(validator=attr.validators.instance_of(bool), default=False)

    def process(self):
        if len(self.column_name) == 0:
            raise ValueError("The name of the column to explode is empty.")

        selected_field_names: typing.List[str] = [ ]
        field_name: str

        if self.type_names is not None:
            if self.verbose:
                print("Validate the names of the data types to extract.", file=self.error_file, flush=True)
            type_name: str
            for type_name in self.type_names:
                if type_name not in KgtkValueFields.DEFAULT_DATA_TYPE_FIELDS:
                    raise ValueError("Unknown data type name '%s'." % type_name)
                # Merge this KGTK data type's fields into the list of selected fields:
                for field_name in KgtkValueFields.DEFAULT_DATA_TYPE_FIELDS[type_name]:
                    if field_name not in selected_field_names:
                        selected_field_names.append(field_name)

        if self.field_names is not None:
            # Forget the fields selected above, choose these instead:
            selected_field_names = [ ]
            if self.verbose:
                print("Validate the names of the fields to extract.", file=self.error_file, flush=True)
            for field_name in self.field_names:
                if field_name not in KgtkValueFields.FIELD_NAMES:
                    raise ValueError("Unknown field name '%s'." % field_name)
                # Merge this field into the list of selected fields:
                if field_name not in selected_field_names:
                    selected_field_names.append(field_name)

        if len(selected_field_names) == 0:
            raise ValueError("The list of fields to explode is empty.")

        # Open the input file.
        if self.verbose:
            if self.input_file_path is not None:
                print("Opening the input file: %s" % self.input_file_path, file=self.error_file, flush=True)
            else:
                print("Reading the input data from stdin", file=self.error_file, flush=True)

        kr: KgtkReader =  KgtkReader.open(self.input_file_path,
                                          error_file=self.error_file,
                                          options=self.reader_options,
                                          value_options = self.value_options,
                                          verbose=self.verbose,
                                          very_verbose=self.very_verbose,
        )

        if self.verbose:
            print("Check that the source column '%s' is present." % self.column_name, file=self.error_file, flush=True)
        if self.column_name not in kr.column_name_map:
            raise ValueError("Column name '%s' not found in the input file." % self.column_name)
        column_idx: int = kr.column_name_map[self.column_name]

        if self.verbose:
            print("Build the map of exploded columns and list of new column names", file=self.error_file, flush=True)
        explosion: typing.MutableMapping[str, idx] = { }
        column_names: typing.List[str] = kr.column_names.copy()
        for field_name in selected_field_names:
            exploded_name: str = self.prefix + field_name
            if self.verbose:
                print("Field '%s' becomes '%s'" % (field_name, exploded_name), file=self.error_file, flush=True)
            if exploded_name in explosion:
                raise ValueError("Field name '%s' is duplicated in the field list.")
            if exploded_name in kr.column_names:
                if self.overwrite_columns:
                    existing_idx = kr.column_name_map[exploded_name]
                    explosion[field_name] = existing_idx
                    if self.verbose:
                        print("Field '%s' is overwriting existing column '%s' (idx=%d)" % (field_name, exploded_name, existing_idx),
                              file=self.error_file, flush=True)
                else:
                    raise ValueError("Exploded column '%s' already exists and not allowed to overwrite" % exploded_name)
            else:
                column_names.append(exploded_name)
                exploded_idx: int = len(column_names) - 1
                explosion[field_name] = exploded_idx
                if self.verbose:
                    print("Field '%s' becomes new column '%s' (idx=%d)" % (field_name, exploded_name, exploded_idx), file=self.error_file, flush=True)
        new_column_count: int = len(column_names) - kr.column_count
        if self.verbose:
            print("%d columns + %d columns = %d columns" % (kr.column_count, new_column_count, len(column_names)))
            print("Explosion length: %d" % len(explosion))
                
        # Open the output file.
        ew: KgtkWriter = KgtkWriter.open(column_names,
                                         self.output_file_path,
                                         mode=kr.mode,
                                         output_format=self.output_format,
                                         require_all_columns=False,
                                         prohibit_extra_columns=True,
                                         fill_missing_columns=True,
                                         gzip_in_parallel=False,
                                         verbose=self.verbose,
                                         very_verbose=self.very_verbose)        
        
        if self.verbose:
            print("Expanding records from %s" % self.input_file_path, file=self.error_file, flush=True)
        input_line_count: int = 0
        output_line_count: int = 0;

        row: typing.List[str]
        for row in kr:
            input_line_count += 1

            # Parse the value for the colummn being exploded:
            item_to_explode: str = row[column_idx]
            value: KgtkValue = KgtkValue(item_to_explode, options=self.value_options, parse_fields=True)
            value.validate()
            if not value.is_valid():
                if self.verbose:
                    print("Not exploding invalid item '%s' in input line %d" % (item_to_explode, input_line_count), file=self.error_file, flush=True)
                ew.write(row) # This will be filled to the proper length
                output_line_count += 1
                continue

            if self.expand_list and value.is_list():
                if self.verbose:
                    print("Expanding a list: '%s'" % item_to_explode, file=self.error_file, flush=True)
                subvalue: KgtkValue
                for subvalue in value.get_list_items():
                    if self.very_verbose:
                        print("Exploding '%s'" % subvalue.value)
                    ew.write(self.explode(subvalue, row, explosion, new_column_count))
                    output_line_count += 1
            else:
                if self.very_verbose:
                    print("Exploding '%s'" % value.value)
                ew.write(self.explode(value, row, explosion, new_column_count))
                output_line_count += 1

        if self.verbose:
            print("Read %d records, wrote %d records." % (input_line_count, output_line_count), file=self.error_file, flush=True)
        
        ew.close()

    def explode(self, value: KgtkValue, row: typing.List[str], explosion: typing.Mapping[str, int], new_column_count: int)->typing.List[str]:
        newrow: typing.List[str] = row.copy()
        if new_column_count > 0:
            # Would it be better to do:
            #
            # if new_column_count > 0:
            #     newrow.extend(["] * new_column_count)
            i: int
            for i in range(new_column_count):
                newrow.append("")
        field_map: typing.Mapping[str, typing.Union[str, int, float, bool]] = value.get_field_map()
        field_name: str
        idx: int
        for field_name, idx in explosion.items():
            if field_name in field_map:
                newvalue: str
                if KgtkValueFields.FIELD_NAME_FORMATS[field_name] == "str":
                    # Format this as a KGTK string.
                    newvalue = '"' + str(field_map[field_name]) + '"'
                else:
                    # Convert everything else to a KGTK number or symbol
                    newvalue = str(field_map[field_name])
                newrow[idx] = newvalue
            else:
                newrow[idx] = "" # In case we are overwriting an existing column.
        return newrow
            

def main():
    """
    Test the KGTK ifempty processor.
    """
    parser: ArgumentParser = ArgumentParser()

    parser.add_argument(dest="input_file_path", help="The KGTK file with the input data. (default=%(default)s)", type=Path, nargs="?", default="-")

    parser.add_argument(      "--column", dest="column_name", help="The name of the column to explode. (default=%(default)s).", default="node2")

    fgroup: ArgumentParser = parser.add_mutually_exclusive_group()

    fgroup.add_argument(      "--types", dest="type_names", nargs='*',
                               help="The KGTK data types for which fields should be exploded. (default=%(default)s).",
                               choices=KgtkFormat.DataType.choices(),
                               default=KgtkFormat.DataType.choices())

    fgroup.add_argument(      "--fields", dest="field_names",  nargs='*',
                              help="The names of the fields to extract (overrides --types). (default=%(default)s).",
                              choices=KgtkValueFields.FIELD_NAMES)

    parser.add_argument("-o", "--output-file", dest="output_file_path", help="The KGTK file to write (default=%(default)s).", type=Path, default="-")
    
    parser.add_argument(      "--prefix", dest="prefix", help="The prefix for exploded column names. (default=%(default)s).", default="node2;kgtk:")

    parser.add_argument(      "--overwrite", dest="overwrite_columns",
                              help="Indicate that it is OK to overwrite existing columns. (default=%(default)s).",
                              type=optional_bool, nargs='?', const=True, default=False)

    parser.add_argument(      "--expand", dest="expand_list",
                              help="Expand the source column if it contains a list, else fail. (default=%(default)s).",
                              type=optional_bool, nargs='?', const=True, default=False)

    KgtkReader.add_debug_arguments(parser)
    KgtkReaderOptions.add_arguments(parser, mode_options=True)
    KgtkValueOptions.add_arguments(parser)

    args: Namespace = parser.parse_args()

    error_file: typing.TextIO = sys.stdout if args.errors_to_stdout else sys.stderr

    # Build the option structures.                                                                                                                          
    reader_options: KgtkReaderOptions = KgtkReaderOptions.from_args(args)
    value_options: KgtkValueOptions = KgtkValueOptions.from_args(args)

   # Show the final option structures for debugging and documentation.                                                                                             
    if args.show_options:
        # TODO: show ifempty-specific options.
        print("input: %s" % str(args.input_file_path), file=error_file, flush=True)
        print("--column %s" % args.column_name, file=error_file, flush=True)
        print("--prefix %s" % args.prefix, file=error_file, flush=True)
        print("--overwrite %s" % str(args.overwrite_columns), file=error_file, flush=True)
        print("--expand %s" % str(args.expand_list), file=error_file, flush=True)
        if args.field_names is not None:
            print("--fields %s" % " ".join(args.field_names), file=error_file, flush=True)
        if args.type_names is not None:
            print("--types %s" % " ".join(args.type_names), file=error_file, flush=True)
        print("--output-file=%s" % str(args.output_file_path))
        reader_options.show(out=error_file)
        value_options.show(out=error_file)

    ex: KgtkExplode = KgtkExplode(
        input_file_path=args.input_file_path,
        column_name=args.column_name,
        prefix=args.prefix,
        field_names=args.field_names,
        type_names=args.type_names,
        overwrite_columns=args.overwrite_columns,
        expand_list=args.expand_list,
        output_file_path=args.output_file_path,
        reader_options=reader_options,
        value_options=value_options,
        error_file=error_file,
        verbose=args.verbose,
        very_verbose=args.very_verbose)

    ex.process()

if __name__ == "__main__":
    main()
